#ifndef SETPLAY_MESSAGE_PARSER_H
#define SETPLAY_MESSAGE_PARSER_H

#include <rcsc/common/say_message_parser.h>
#include <rcsc/game_time.h>
#include <boost/shared_ptr.hpp>

#include <setplay/communication/messages.h>

#include <setplay/runtime/setplay_communication.h>
#include <setplay/runtime/setplay_database.h>


namespace fcportugal
{
    class AudioMemory;

}

class SetplayMessageParser
        : public rcsc::SayMessageParser {

    public:
        SetplayMessageParser( setplay::SetplayCommunication* setplay_communication,
                              setplay::SetplayDatabase* setplay_database)
            : M_setplay_communication( setplay_communication )
            , M_setplay_database( setplay_database)
        {

        }

        virtual ~SetplayMessageParser() { }

        virtual
        char header() const = 0;

        virtual
        int parse( const int sender,
                   const double & dir,
                   const char * msg,
                   const rcsc::GameTime & current ) = 0;
protected:

        unsigned int parseSetplayBase( const char* msg,
                                       unsigned int& setplay_id,
                                       bool& inverted );

        unsigned int parseSetplayArguments( const char* msg,
                                            setplay::Setplay& setplay,
                                            setplay::IMessage& smsg );

    //! pointer to setplay communication manager
    setplay::SetplayCommunication* M_setplay_communication;
    //! pointer to setplay mediator (contains database interface)
    setplay::SetplayDatabase* M_setplay_database;

};

// j, k, l, m, n  => not used in say_message_parser taken

/*!
  \class StepTransitionMessageParser
  \brief Setplay step transition information message parser

  format:
  "j<curr_step:1><next_step:1>"
  the length of message == 3  format:
 */
class StepTransitionMessageParser
        : public SetplayMessageParser
{
    public:
        StepTransitionMessageParser( setplay::SetplayCommunication* setplay_communication,
                                     setplay::SetplayDatabase* setplay_database );
        virtual ~StepTransitionMessageParser();

        static
        char sheader() { return 'j'; }

        char header() const { return sheader(); }

        /*!
          \brief get the length of this message.
          \return the length of encoded message
        */
        static
        int slength() { return 3; }

        int parse( const int sender,
                   const double & dir,
                   const char * msg,
                   const rcsc::GameTime & current );

private:
        unsigned int parseSteps( const char* msg,
                                 setplay::StepTransitionMessage& smsg ) const;
};


/*!
  \class StepEnterMessageParser
  \brief Setplay step enter information message parser

  format:
  "j<curr_step:1><next_step:1>"
  the length of message == 3  format:
 */
class StepEnterMessageParser
        : public SetplayMessageParser
{
    public:
        StepEnterMessageParser( setplay::SetplayCommunication* setplay_communication,
                                     setplay::SetplayDatabase* setplay_database );
        virtual ~StepEnterMessageParser();

        static
        char sheader() { return 'k'; }

        char header() const { return sheader(); }

        /*!
          \brief get the length of this message.
          \return the length of encoded message
        */
        static
        int slength() { return 3; }

        int parse( const int sender,
                   const double & dir,
                   const char * msg,
                   const rcsc::GameTime & current );
};


/*!
  \class SetplayStartMessageParser
  \brief Setplay start information message parser

  format:
  "l<unum_1:1>...<unum_n:1>"
  the length of message == 1 + len(args)  format:
 */
class SetplayStartMessageParser
        : public SetplayMessageParser
{
public:
    SetplayStartMessageParser( setplay::SetplayCommunication* setplay_communication,
                               setplay::SetplayDatabase* setplay_database);
    virtual ~SetplayStartMessageParser();

    static
    char sheader() { return 'l'; }

    char header() const { return sheader(); }

    int parse( const int sender,
               const double & dir,
               const char * msg,
               const rcsc::GameTime & current );
};


/*!
  \class AbortMessageParser
  \brief Setplay start information message parser

  format:
  "m"
  the length of message == 1 format:
 */
class SetplayAbortMessageParser
        : public SetplayMessageParser
{
public:
    SetplayAbortMessageParser( setplay::SetplayCommunication* setplay_communication,
                               setplay::SetplayDatabase* setplay_database ):
        SetplayMessageParser(setplay_communication, setplay_database)
    {

    }

    static
    char sheader() { return 'm'; }

    char header() const { return sheader(); }

    static
    int slength() { return 1; }

    int parse( const int sender,
               const double & dir,
               const char * msg,
               const rcsc::GameTime & current );
};

/*!
  \class FinishMessageParser
  \brief Setplay start information message parser

  format:
  "n"
  the length of message == 1 format:
 */
 class SetplayFinishMessageParser
        : public SetplayMessageParser
{
public:
    SetplayFinishMessageParser( setplay::SetplayCommunication* setplay_communication,
                                setplay::SetplayDatabase* setplay_database ):
        SetplayMessageParser(setplay_communication, setplay_database)
    {

    }

    static
    char sheader() { return 'n'; }

    char header() const { return sheader(); }

    static
    int slength() { return 1; }

    int parse( const int sender,
               const double & ,
               const char * msg,
               const rcsc::GameTime & current );
};

#endif // SETPLAY_MESSAGE_PARSER_H
