#include "world_model.h"

#include <utility>
#include <sstream>

#include "../strategy.h"

#include <functional>
#include <algorithm>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom.h>
#include <rcsc/game_mode.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/player_object.h>

#include <setplay/model/geom.h>

using namespace rcsc;

namespace soccer {

WorldModel::WorldModel( const rcsc::WorldModel *world )
    : M_world( world )
      // Regions
    , M_field( Vector2D(-ServerParam::i().pitchHalfLength(), -ServerParam::i().pitchHalfWidth()),
               Size2D(ServerParam::i().pitchLength(), ServerParam::i().pitchWidth()))
    , M_left_goalie_area( Vector2D( -ServerParam::i().pitchHalfLength(), -ServerParam::i().goalAreaHalfWidth() ),
                          Size2D( ServerParam::i().goalAreaLength(), ServerParam::i().goalAreaWidth()))
    , M_right_goalie_area( Vector2D( ServerParam::i().pitchHalfLength() - ServerParam::i().goalAreaLength(), -ServerParam::i().goalAreaHalfWidth() ),
                           Size2D( ServerParam::i().goalAreaLength(), ServerParam::i().goalAreaWidth()))
    , M_left_penalty_area( Vector2D( -ServerParam::i().pitchHalfLength(), -ServerParam::i().penaltyAreaHalfWidth() ),
                           Size2D( ServerParam::i().penaltyAreaLength(), ServerParam::i().penaltyAreaWidth()))
    , M_right_penalty_area( Vector2D( ServerParam::i().pitchHalfLength() - ServerParam::i().penaltyAreaLength(), -ServerParam::i().penaltyAreaHalfWidth() ),
                            Size2D( ServerParam::i().penaltyAreaLength(), ServerParam::i().penaltyAreaWidth()))
    , M_left_midfield_area( Vector2D( -ServerParam::i().pitchHalfLength(), -ServerParam::i().pitchHalfWidth()),
                            Size2D(ServerParam::i().pitchHalfLength(), ServerParam::i().pitchHalfWidth()))
    , M_right_midfield_area( Vector2D(0, -ServerParam::i().pitchHalfWidth()),
                             Size2D(ServerParam::i().pitchHalfLength(), ServerParam::i().pitchHalfWidth()))

    , M_left_end_line( -ServerParam::i().pitchHalfLength())
    , M_right_end_line( ServerParam::i().pitchHalfLength())
    , M_left_goal_line( -ServerParam::i().pitchHalfLength())
    , M_right_goal_line( ServerParam::i().pitchHalfLength())
    , M_current_game_mode( GameSituation(GameSituation::BeforeKickOff) )
    , M_previous_game_mode( M_current_game_mode )
    , M_teammates_from_own_goal_cmp( world->ourPlayers(),
                                     ServerParam::i().ourGoalArea().center() )
    , M_opponents_from_own_goal_cmp( world->theirPlayers(),
                                     ServerParam::i().theirGoalArea().center() )
    , M_soccer_moment( SM_GAME_STOPPED_THEIR_POSSESION )
{

    // Initialize players from own goal
    // Initialize players from opp goal
    for ( AbstractPlayerCont::const_iterator itp = M_world->ourPlayers().begin();
          itp != M_world->ourPlayers().end(); itp++)
    {
        M_our_players_from_own_goal.push_back( (*itp)->unum() );
    }

    for ( AbstractPlayerCont::const_iterator itp = M_world->theirPlayers().begin();
          itp != M_world->theirPlayers().end(); itp++)
    {
        M_opp_players_from_own_goal.push_back( (*itp)->unum() );
    }

    initPlayModes();
}

void WorldModel::update( )
{
    // dlog.addText( Logger::WORLD, __FILE__ " Updating our players current position for strategy");
    // updatePlayerPositions(M_our_players_positions, M_world_model->ourPlayers());
    // dlog.addText( Logger::WORLD, __FILE__ " Updating opp players current position for strategy");
    // updatePlayerPositions(M_opp_players_positions, M_world_model->theirPlayers());

    /// Update team side
    /* if ( M_world->ourSide() == LEFT )
    {
        M_our_side = rcsc::LEFT; M_opp_side = TS_RIGHT;
    }
    else
    {
        M_our_side = TS_RIGHT; M_opp_side = rcsc::LEFT;
    } */

    updateGameMode( );
    updatePlayersCalculus();
    updateSoccerMoment();

    updateScoreAssessment();
    updateGamePace();

    dlog.addText( Logger::PLAN,
                  __FILE__ " WorldModel: selfReachCycle=%d=%s teammateReachCycle=%d=%s oppReachCycle=%d=%s BallPosValid=%s",
                  M_world->interceptTable()->selfReachCycle(),
                  M_world->self().isKickable() ? "TRUE" : "FALSE",
                  M_world->interceptTable()->teammateReachCycle(),
                  M_world->existKickableTeammate() ? "TRUE" : "FALSE",
                  M_world->interceptTable()->opponentReachCycle(),
                  M_world->existKickableOpponent()? "TRUE" : "FALSE",
                  M_world->ball().posValid() ? "TRUE" : "FALSE" );

    // Update players behind ball from both teams
    // updatePlayersRelativeToBall();

    // updateGameSituation();

    // updatePlayersOffside();

}


/// Override IWorldModel
rcsc::SideID WorldModel::side( const TeamAffiliation &affiliation) const
{
    assert ( affiliation != TA_UNKNOWN );
    return affiliation == TA_OUR
            ? M_world->ourSide()
            : M_world->theirSide();
}

const rcsc::AbstractPlayerObject* WorldModel::getAbstractPlayerObject( const PlayerAffiliation& player ) const
{
    assert ( player.affiliation() != TA_UNKNOWN );
    const rcsc::AbstractPlayerObject* aplayer = player.affiliation() == TA_OUR
                                                ?  M_world->ourPlayer( player.number() )
                                                : M_world->theirPlayer( player.number() );
    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " WorldModel: Could not get a reference to player %s #%d",
                      player.affiliation() == TA_OUR ? "our" : "opp",
                      player.number() );
    }

    return aplayer;
}


unsigned int WorldModel::currentTime() const
{
    return M_world->time().cycle() + M_world->time().stopped();
}

unsigned int WorldModel::remainingTimeInPeriod() const
{
    switch( M_time_period )
    {
        case TP_FIRST_HALF:
            return ServerParam::i().halfTime() - M_current_time;
        case TP_SECOND_HALF:
            return ServerParam::i().halfTime() * 2 - M_current_time;
        case TP_EXTRA_TIME_FIRST_HALF:
            return ( ServerParam::i().halfTime() * 2 + ServerParam::i().extraHalfTime() ) - M_current_time;
        case TP_EXTRA_TIME_SECOND_HALF:
            return ( ServerParam::i().halfTime() * 2 + ServerParam::i().extraHalfTime() * 2 ) - M_current_time;
        default:
            assert( false );
    }
}

GameSituation WorldModel::gameSituation() const
{
    return M_current_game_mode.gameSituation();
}

// Game stopped?
GameMode WorldModel::currentGameMode() const
{
    return M_current_game_mode;
}

SoccerMomentAffiliation WorldModel::soccerMoment() const
{
    return M_soccer_moment;
}

unsigned int WorldModel::ballInterceptionTime( const TeamAffiliation& team,
                                               const Unum& unum) const
{
    assert( false );
}

unsigned int WorldModel::playerInterceptionTime( const TeamAffiliation& team1,
                                                 const Unum& unum1,
                                                 const TeamAffiliation& team2,
                                                 const Unum& unum2) const
{
    assert( false );
}

TimePeriod WorldModel::timePeriod() const
{
    return M_time_period;
}


/// COMMON REGIONS
double WorldModel::defenseLine(TeamAffiliation team_affiliation) const
{
    return team_affiliation == TA_OUR
            ? M_our_defense_line
            : M_opp_defense_line;
}

double WorldModel::offenseLine(TeamAffiliation team_affiliation) const
{
    return team_affiliation == TA_OUR
            ? M_our_offense_line
            : M_opp_offense_line;
}

double WorldModel::goalLine(TeamAffiliation team_affiliation) const
{
    return side(team_affiliation) == rcsc::LEFT
            ? M_left_goal_line
            : M_right_goal_line;
}

double WorldModel::endLine(TeamAffiliation team_affiliation) const
{
    return side(team_affiliation) == rcsc::LEFT
            ? M_left_end_line
            : M_right_end_line;
}

rcsc::Rect2D WorldModel::field( ) const
{
    return M_field;
}

rcsc::Rect2D WorldModel::goalieArea(TeamAffiliation team_affiliation) const
{
    return side(team_affiliation) == rcsc::LEFT
            ? M_left_goalie_area
            : M_right_goalie_area;
}

rcsc::Rect2D WorldModel::penaltyArea(TeamAffiliation team_affiliation) const
{
    return side(team_affiliation) == rcsc::LEFT
            ? M_left_penalty_area
            : M_right_penalty_area;
}

rcsc::Rect2D WorldModel::midfieldArea(TeamAffiliation team_affiliation) const
{
    return side(team_affiliation) == rcsc::LEFT
            ? M_left_midfield_area
            : M_right_midfield_area;
}


rcsc::Vector2D WorldModel::ballPosition( ) const
{
    if ( M_current_game_mode.gameSituation().type() == GameSituation::Goal )
    {
        return Vector2D(0.0, 0.0); // HACK For player teletransport
    }
    return M_world->ball().pos();
}

double WorldModel::ballSpeed() const
{
    return M_world->ball().vel().r();
}

rcsc::AngleDeg WorldModel::ballDirection() const
{
    return M_world->ball().vel().dir();
}

/// PLAYER INFORMATION
unsigned int WorldModel::self() const
{
    return M_world->self().unum();
}

unsigned int WorldModel::goalie( TeamAffiliation team_affiliation ) const
{
    return team_affiliation == TA_OUR
            ? M_world->ourGoalieUnum()
            : M_world->theirGoalieUnum();
}

rcsc::Vector2D WorldModel::playerPosition( const TeamAffiliation& team,
                                           const Unum& unum ) const
{
    assert( team != TA_UNKNOWN );

    const AbstractPlayerObject* aplayer = getAbstractPlayerObject( PlayerAffiliation( unum, team) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " WorldModel: Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum );
        return Vector2D::INVALIDATED;
    }

    return aplayer->pos();
}

double WorldModel::playerSpeed( const TeamAffiliation& team,
                                const Unum& unum ) const
{
    const AbstractPlayerObject* aplayer = getAbstractPlayerObject( PlayerAffiliation( unum, team ) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum );
        return ServerParam::i().playerAccelMax(); // HACK
    }

    return aplayer->vel().r();
}

rcsc::AngleDeg WorldModel::playerBodyDirection( const TeamAffiliation& team,
                                                const Unum& unum) const
{
    const AbstractPlayerObject* aplayer = getAbstractPlayerObject( PlayerAffiliation( unum, team) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum );
        return AngleDeg(); // HACK To return some direction
    }

    return aplayer->body();
}

int WorldModel::playerTimeToReachPosition( const TeamAffiliation &team,
                                           const Unum& unum,
                                           const Vector2D &position ) const
{
    const AbstractPlayerObject* aplayer = getAbstractPlayerObject( PlayerAffiliation(unum, team) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum );
        return 99999; // HACK Return very large number
    }

    return aplayer->playerTypePtr()->cyclesToReachDistance( aplayer->pos().dist( position ) );
}

int WorldModel::playerTimeToReachMaxSpeed( const TeamAffiliation& team,
                                           const Unum& unum ) const
{
    const AbstractPlayerObject* aplayer = getAbstractPlayerObject( PlayerAffiliation( unum, team) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum );
        return 99999; // Return a very large number!!! Use limits
    }

    return aplayer->playerTypePtr()->cyclesToReachMaxSpeed();
}

double WorldModel::playerMaxSpeed( const TeamAffiliation& team,
                                   const Unum& unum ) const
{
    const AbstractPlayerObject* aplayer = getAbstractPlayerObject(  PlayerAffiliation( unum, team) );

    if ( aplayer == NULL )
    {
        dlog.addText( Logger::WORLD,
                      __FILE__ " Could not get a reference to player %s #%d",
                      team == TA_OUR ? "our" : "opp",
                      unum);
        return 99999; // Return a very large number!!! Use limits
    }

    return aplayer->playerTypePtr()->realSpeedMax( );
}

vector<unsigned int> WorldModel::activePlayers( TeamAffiliation team_affiliation ) const
{
    std::vector<unsigned int> active_players;

    for ( AbstractPlayerCont::const_iterator itp = M_world->ourPlayers().begin();
          itp != M_world->ourPlayers().end() ; itp++)
    {
        const AbstractPlayerObject* player = *itp;
        if ( player->unum() < 0 )
        {
            dlog.addText( Logger::PLAN,
                          __FILE__ " WorldModel: Player with illegal number %d", player->unum() );
            continue;
        }
        active_players.push_back( (*itp)->unum() );
    }
    return active_players;
}

PlayerAffiliation WorldModel::playerFastestTo( const rcsc::Vector2D& position )
{
    assert(false);
}

PlayerAffiliation WorldModel::playerFastestTo( const rcsc::Region2D& region )
{
    assert(false);
}

PlayerAffiliation WorldModel::playerFastestTo( const TeamAffiliation& team,
                                               const Unum& unum )
{
    assert(false);
}

PlayerAffiliation WorldModel::playerFastestToBall( )
{
    if ( M_world->interceptTable()->teammateReachCycle() <= M_world->interceptTable()->opponentReachCycle() )
    {
        return PlayerAffiliation( M_world->interceptTable()->fastestTeammate()->unum(),
                                          TA_OUR );
    }
    return PlayerAffiliation( M_world->interceptTable()->fastestOpponent()->unum(),
                                      TA_OPP );
}


unsigned int WorldModel::teammateFastestTo( const rcsc::Vector2D& position )
{
    assert(false);
}

unsigned int WorldModel::teammateFastestTo( const rcsc::Region2D& region )
{
    assert(false);
}

unsigned int WorldModel::teammateFastestTo( const TeamAffiliation &team, const Unum &unum )
{
    /// Check the player fastest to another player
    assert(false);
}

unsigned int WorldModel::teammateFastestToBall( )
{
    return M_world->interceptTable()->fastestTeammate()->unum();
}

unsigned int WorldModel::opponentFastestTo( const rcsc::Vector2D& position )
{
    assert(false);
}

unsigned int WorldModel::opponentFastestTo( const rcsc::Region2D& region )
{
    assert(false);
}

unsigned int WorldModel::opponentFastestTo( const TeamAffiliation &team, const Unum &unum )
{
    assert(false);
}

unsigned int WorldModel::opponentFastestToBall( )
{
    return M_world->interceptTable()->fastestOpponent()->unum();
}


std::vector<unsigned int> WorldModel::teammatesFromPlayer( const TeamAffiliation& team_affiliation,
                                                           const Unum& number ) const
{
    assert(false);
}

std::vector<unsigned int> WorldModel::opponentsFromPlayer( const TeamAffiliation& team_affiliation,
                                                           const Unum& number) const
{
    assert(false);
}

vector<unsigned int> WorldModel::playersFromBall( TeamAffiliation team_affiliation ) const
{
    if ( team_affiliation == TA_OUR )
    {
        return M_our_players_from_ball;
    }
    return M_opp_players_from_ball;
}

vector<unsigned int> WorldModel::playersFromGoal( TeamAffiliation team_affiliation,
                                                  TeamAffiliation goal_affiliation) const
{
    if ( team_affiliation == TA_OUR )
    {
        return M_our_players_from_own_goal;
    }
    return M_opp_players_from_own_goal;
}

vector<unsigned int> WorldModel::playersBehindBall( TeamAffiliation team_affiliation) const
{
    if ( team_affiliation == TA_OUR )
    {
        return M_our_players_behind_ball;
    }
    return M_opp_players_behind_ball;
}

vector<unsigned int> WorldModel::playersInFrontOfBall( TeamAffiliation team_affiliation) const
{
    if ( team_affiliation == TA_OUR )
    {
        return M_our_players_in_front_of_ball;
    }
    return M_opp_players_in_front_of_ball;
}

PlayerAffiliation WorldModel::lastKicker() const
{
    // Unknown player
}

TeamAffiliation WorldModel::teamInPossession() const
{
    if ( M_world->gameMode().isOurSetPlay( M_world->ourSide() ) )
    {
        return TA_OUR;
    }

    if ( M_world->gameMode().isTheirSetPlay( M_world->ourSide() ) )
    {
        return TA_OPP;
    }

    const int reach_cycle_thr = 5;
    /// Assuming play-on mode
    if ( M_world->gameMode().type() == rcsc::GameMode::PlayOn )
    {
        dlog.addText( Logger::PLAN,
                      __FILE__ " WorldModel: Reach cycles Self=%d Teammate=%d Opp=%d Thr=%d ",
                      M_world->interceptTable()->selfReachCycle(),
                      M_world->interceptTable()->teammateReachCycle(),
                      M_world->interceptTable()->opponentReachCycle(),
                      reach_cycle_thr);
        if ( ( M_world->interceptTable()->selfReachCycle() <= M_world->interceptTable()->opponentReachCycle() + reach_cycle_thr ) ||
             ( M_world->interceptTable()->teammateReachCycle() <= M_world->interceptTable()->opponentReachCycle() + reach_cycle_thr) )
        {
            return TA_OUR;
        }
        return TA_OPP;
    }

    stringstream ss;
    M_world->gameMode().print(ss);
    dlog.addText( Logger::PLAN,
                  __FILE__ " WorldModel: teamInPossession() => Game mode %s not considered. Returning OPP by default",
                  ss.str().c_str());
    return TA_OPP;
}

unsigned int WorldModel::goalsScored(TeamAffiliation team_affiliation) const
{
    assert ( team_affiliation != TA_UNKNOWN );
    if ( team_affiliation == TA_OUR )
    {
        return M_world->ourSide() == LEFT
                ? M_world->gameMode().scoreLeft()
                : M_world->gameMode().scoreRight();
    }
    return M_world->theirSide() == LEFT
            ? M_world->gameMode().scoreLeft()
            : M_world->gameMode().scoreRight();
}

std::vector<unsigned int> WorldModel::playersOffside(TeamAffiliation team_affiliation) const
{
    assert ( team_affiliation != TA_UNKNOWN );

    return team_affiliation == TA_OUR
            ? M_our_players_offside
            : M_opp_players_offside;
}

std::vector<unsigned int> WorldModel::playersNotOffside(TeamAffiliation team_affiliation) const
{
    assert ( team_affiliation != TA_UNKNOWN );

    return team_affiliation == TA_OUR
            ? M_our_players_non_offside
            : M_opp_players_non_offside;
}


void WorldModel::setSoccerMoment( const SoccerMomentAffiliation &soccer_moment )
{
    if ( M_soccer_moment == soccer_moment ) { return; }

    /* dlog.addText( Logger::PLAN,
                  __FILE__ " WorldModel: SoccerMomentChanged from %s to %s",
                  soccer::soccerMoment( M_soccer_moment ).c_str(),
                  soccer::soccerMoment( soccer_moment ).c_str());
    std::cerr << M_world->teamName() << "-" << M_world->self().unum()
              << " SoccerMomentChanged from " << soccer::soccerMoment( M_soccer_moment )
              << " to " << soccer::soccerMoment( soccer_moment )
              << std::endl; */
    M_soccer_moment = soccer_moment;
}

//
// World Model updates
//
void WorldModel::updateSoccerMoment()
{
    if ( M_world->gameMode().isOurSetPlay( M_world->ourSide() ) )
    {
        setSoccerMoment( SM_GAME_STOPPED_OUR_POSSESION );
        return;
    }

    if ( M_world->gameMode().isTheirSetPlay( M_world->ourSide() ) )
    {
        setSoccerMoment( SM_GAME_STOPPED_THEIR_POSSESION );
        return;
    }

    /// PlayOn
    switch ( teamInPossession() )
    {
        case TA_OUR:
        {
            if( M_soccer_moment == SM_GAME_STOPPED_THEIR_POSSESION )
            {
                setSoccerMoment( SM_OUR_POSSESION );
                return;
            }

            if ( M_soccer_moment != SM_OUR_POSSESION )
            {
                setSoccerMoment( ( M_world->self().isKickable() || M_world->existKickableTeammate() )
                                    ? SM_OUR_POSSESION
                                    : SM_REGAINING_POSSESION );
            }
        }
        break;
        case TA_OPP:
        {
            if( M_soccer_moment == SM_GAME_STOPPED_OUR_POSSESION )
            {
                setSoccerMoment( SM_THEIR_POSSESION );
                return;
            }

            if ( M_soccer_moment != SM_THEIR_POSSESION )
            {
                setSoccerMoment( (M_world->existKickableOpponent() )
                                    ? SM_THEIR_POSSESION
                                    : SM_LOSING_POSSESION );
            }
        }
        break;
    default:
        dlog.addText( Logger::PLAN,
                      __FILE__ " WorldModel: updateSoccerMoment: Could not determine teamInPossession");
        break;
    }
}

void WorldModel::updateScore( )
{
    if ( M_world->ourSide() == LEFT )
    {
        M_our_score = M_world->gameMode().scoreLeft();
        M_opp_score = M_world->gameMode().scoreRight();
    }
    else
    {
        M_our_score = M_world->gameMode().scoreRight();
        M_opp_score = M_world->gameMode().scoreLeft();
    }


    M_relative_score = M_our_score - M_opp_score;
}

void WorldModel::updatePlayerOffside(vector<unsigned int>& players_offside,
                                      vector<unsigned int>& players_not_offside,
                                      double offside_line,
                                      const AbstractPlayerObject& player  )
{
    if ( abs(player.pos().x) < offside_line )
    {
        players_offside.push_back( player.unum() );
        return;
    }
    players_not_offside.push_back( player.unum() );
}

void WorldModel::updatePlayerRelativeToBallX( vector<unsigned int>& players_behind_ball,
                                               vector<unsigned int>& players_in_front_ball,
                                               const AbstractPlayerObject& player  )
{
    if ( abs(player.pos().x) <= abs(M_world->ball().pos().x) )
    {
        players_behind_ball.push_back( player.unum() );
        return;
    }

    players_in_front_ball.push_back( player.unum() );
}

void WorldModel::updatePlayersFromBall( vector<unsigned int>& player_ids_from_ball,
                                        const PlayerPtrCont &players_from_ball)
{
    for ( PlayerPtrCont::const_iterator itp = players_from_ball.begin();
          itp != players_from_ball.end(); itp++)
    {
        player_ids_from_ball.push_back( (*itp)->unum() );
    }
}

void WorldModel::updatePlayersCalculus(  )
{
    // Check which opponents are in an offside even if they're not attacking
    M_our_players_offside.clear();
    M_opp_players_offside.clear();
    M_our_players_non_offside.clear();
    M_opp_players_non_offside.clear();
    M_our_players_behind_ball.clear();
    M_opp_players_behind_ball.clear();
    M_our_players_in_front_of_ball.clear();
    M_opp_players_in_front_of_ball.clear();

    for ( AbstractPlayerCont::const_iterator itp = M_world->ourPlayers().begin();
          itp != M_world->ourPlayers().end(); itp++ )
    {
        const AbstractPlayerObject* player = *itp;
        if ( player->side() == M_world->ourSide() )
        {
            updatePlayerRelativeToBallX( M_our_players_behind_ball,
                                          M_our_players_in_front_of_ball,
                                          *player);
            updatePlayerOffside( M_our_players_offside,
                                 M_our_players_non_offside,
                                 abs(M_world->offsideLineX()),
                                 *player);

            return;
        }

        updatePlayerRelativeToBallX( M_opp_players_behind_ball,
                                      M_opp_players_in_front_of_ball,
                                      *player);

        updatePlayerOffside( M_opp_players_offside,
                              M_opp_players_non_offside,
                              abs(M_world->ourDefenseLineX()),
                              *player);
    }

    M_our_players_from_ball.clear();
    M_opp_players_from_ball.clear();

    updatePlayersFromBall( M_our_players_from_ball,
                           M_world->teammatesFromBall());
    updatePlayersFromBall( M_opp_players_from_ball,
                           M_world->opponentsFromBall());

    // Sort teammates based on their distance to own goal
    sort( M_our_players_from_own_goal.begin(),
          M_our_players_from_own_goal.end(),
          M_teammates_from_own_goal_cmp );

    // Sort opponents based on their distance to own goal
    sort( M_opp_players_from_own_goal.begin(),
          M_opp_players_from_own_goal.end(),
          M_opponents_from_own_goal_cmp);

}

void WorldModel::updateGameMode( )
{
    /*if ( M_world_model->gameMode().getServerPlayMode() == M_last_game_mode.getServerPlayMode() )
    {
        return; // No change in play-mode
    }

    if ( M_last_game_mode.getServerPlayMode() == soccer::GameSituation::BeforeKickOff )
    {
        return;
    } */

    /*dlog.addText( Logger::WORLD,
                 __FILE__ " LastPMode(%d,%s), CurrPMode(%d,%s)",
                 M_last_game_mode.getServerPlayMode(),
                 S_play_mode_strings[M_last_game_mode.getServerPlayMode()],
                 M_world_model->gameMode().getServerPlayMode(),
                 S_play_mode_strings[world_model->gameMode().getServerPlayMode()]);*/

    // M_last_game_mode = M_world_model->gameMode();
    // Innefficient => Should check if playmode changed and also keep a timestamp
    M_previous_game_mode = M_current_game_mode;
    GameSituation game_situation = GameSituation::gameSituation("undefined");
    PlayModeToGameSituationMap::const_iterator it = M_play_modes_map.find(M_world->gameMode().getServerPlayMode());
    stringstream ss;
    M_world->gameMode().print( ss );
    dlog.addText( Logger::PLAN,
                  __FILE__  " WorldModel: Current game mode %s", ss.str().c_str());

    if (it == M_play_modes_map.end())
    {
        if ( M_world->gameMode().type() == rcsc::GameMode::GoalieCatch_ )
        {
            game_situation = M_world->gameMode().side() == M_world->ourSide()
                                ?  GameSituation::gameSituation("goalie_catch_our")
                                : GameSituation::gameSituation("goalie_catch_opp");
            M_current_game_mode = GameMode( GameSituation( game_situation ), M_world->gameMode().time().cycle() );
                                                   // M_world->gameMode().time().cycle() );
            return;
        }

        // Not found...
        dlog.addText( Logger::WORLD,
                     __FILE__ "Invalid playMode %d",
                     M_world->gameMode().getServerPlayMode());
        M_current_game_mode = GameMode( GameSituation(game_situation), M_world->gameMode().time().cycle() );
        return;
    }

    game_situation = it->second;
    M_current_game_mode = GameMode( GameSituation(game_situation), M_world->gameMode().time().cycle() );

    // Must test if it is a goalie catch: it is considered a free kick by the server
    /*playModesMap.insert(pair<PlayMode, PlayMode>(PM_My_Goalie_Free_Kick, GoalieCatch_Our));
    playModesMap.insert(pair<PlayMode, PlayMode>(PM_Their_Goalie_Free_Kick, GoalieCatch_Opp));*/

    //Modos que ficam por testar, por nao haver correspondencia

    // FCPortugal2d:
    //  PM_Half_Time,
    //  PM_Extended_Time,
    //  PM_My_PenSetup,  //LPR 2007
    //  PM_My_PenReady,
    //  PM_My_PenTaken,
    //  PM_My_PenMiss,
    //  PM_My_PenScore,
    //  PM_Their_PenSetup,
    //  PM_Their_PenReady,
    //  PM_Their_PenTaken,
    //  PM_Their_PenMiss,
    //  PM_Their_PenScore
    // Clang:
    //AfterGoal_Our,
    //AfterGoal_Opp
}

/*void WorldModel::updatePlayerPositions(WorldModel::PlayerPositions& player_positions,
                                       const AbstractPlayerCont& players_cont)
{
    player_positions.clear();
    for ( AbstractPlayerCont::const_iterator itp = players_cont.begin(); itp != players_cont.end(); itp++)
    {
        const AbstractPlayerObject* player = *itp;
        if ( player != NULL )
        {
            player_positions[player->unum()] = player->pos();
        }
        else
        {
            dlog.addText( Logger::WORLD, __FILE__ " Could not update a player position for strategy");
        }
    }
}*/

/* void WorldModel::updateGameSituation()
{
    GameSituation previous_game_situation = M_game_situation;
    switch ( M_current_play_mode )
    {
        case GoalKick_Our:
              M_game_situation = soccer::strategy::SI_OGK;
              break;
        case GoalKick_Opp:
              M_game_situation = soccer::strategy::SI_TGK;
              break;
        case FreeKick_Our:
              M_game_situation = soccer::strategy::SI_OFK;
              break;
        case FreeKick_Opp:
              M_game_situation = soccer::strategy::SI_TFK;
              break;
        case Ind_FreeKick_Our:
              M_game_situation = soccer::strategy::SI_OIF;
              break;
        case Ind_FreeKick_Opp:
              M_game_situation = soccer::strategy::SI_TIF;
              break;
        case KickOff_Our:
              M_game_situation = soccer::strategy::SI_OKO;
              break;
        case KickOff_Opp:
              M_game_situation = soccer::strategy::SI_TKO;
              break;
        case KickIn_Our:
              M_game_situation = soccer::strategy::SI_OKI;
              break;
        case KickIn_Opp:
              M_game_situation = soccer::strategy::SI_TKI;
              break;
        case GoalieCatch_Our:
              M_game_situation = soccer::strategy::SI_OGF;
              break;
        case GoalieCatch_Opp:
              M_game_situation = soccer::strategy::SI_TGF;
              break;
        case CornerKick_Our:
              M_game_situation = soccer::strategy::SI_OKI;
              break;
        case CornerKick_Opp:
              M_game_situation = soccer::strategy::SI_TKI;
              break;
        default:
        {
              if (world_model->gameMode().isPenaltyKickMode())
              {
                  M_game_situation  = M_world_model->gameMode().side() == M_world_model->ourSide() ?
                                      soccer::strategy::SI_OPE :
                                      M_game_situation = soccer::strategy::SI_TPE;
              }
              else
              {
              // HACK Returning attack situation by default => DANGER How will this be used
              // cerr << __FILE__ << "Returning attack situation by default " << endl;
                if ( M_our_attackers.size() >= M_their_defenders.size() )
                {
                    dlog.addText( Logger::WORLD,
                                  __FILE__ " Possible CounterAttackSituation: %d vs %d",
                                  M_our_attackers.size(), M_their_attackers.size());
                }

                // Normal attack situation
                M_game_situation = soccer::strategy::SI_ATT;
              }
        }
        break;
    }
    dlog.addText( Logger::WORLD,
                  __FILE__ " PMode(%d), GCatchOur=%d, PrevGameSituation :%d, CurrGameSituation: %d",
                  M_current_play_mode, GoalieCatch_Our, previous_game_situation, M_game_situation);
} */


void WorldModel::initPlayModes( )
{
    // Initialize play-mode maps

    M_play_modes_map[PM_TimeOver] = GameSituation::gameSituation("time_over");
    M_play_modes_map[PM_BeforeKickOff] = GameSituation::gameSituation("before_kick_off");
    M_play_modes_map[PM_PlayOn] = GameSituation(GameSituation::PlayOn);
    M_play_modes_map[PM_Null] = GameSituation::gameSituation("undefined");

    M_play_modes_map[PM_KickOff_Left] = GameSituation::gameSituation("kick_off_our");
    M_play_modes_map[PM_KickOff_Right] = GameSituation::gameSituation("kick_off_opp");

    M_play_modes_map[PM_KickIn_Left] = GameSituation::gameSituation("throw_in_our");
    M_play_modes_map[PM_KickIn_Right] = GameSituation::gameSituation("throw_in_opp");

    M_play_modes_map[PM_FreeKick_Left] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_FreeKick_Right] = GameSituation::gameSituation("free_kick_our");

    M_play_modes_map[PM_IndFreeKick_Left] = GameSituation::gameSituation("indirect_free_kick_our");
    M_play_modes_map[PM_IndFreeKick_Right] = GameSituation::gameSituation("indirect_free_kick_opp");

    M_play_modes_map[PM_CornerKick_Left] = GameSituation::gameSituation("corner_kick_our");
    M_play_modes_map[PM_CornerKick_Right] = GameSituation::gameSituation("corner_kick_opp");

    M_play_modes_map[PM_GoalKick_Left] = GameSituation::gameSituation("goal_kick_our");
    M_play_modes_map[PM_GoalKick_Right] = GameSituation::gameSituation("goal_kick_opp");

    // Fault play-modes Who will own the ball in the next cycle
    M_play_modes_map[PM_OffSide_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_OffSide_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_CatchFault_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_CatchFault_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_Foul_Charge_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_Foul_Charge_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_Foul_MultipleAttacker_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_Foul_MultipleAttacker_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_Foul_Push_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_Foul_Push_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_Back_Pass_Right] = GameSituation::gameSituation("free_kick_our");
    M_play_modes_map[PM_Back_Pass_Left] = GameSituation::gameSituation("free_kick_opp");

    M_play_modes_map[PM_AfterGoal_Left] = GameSituation::gameSituation("goal_our");
    M_play_modes_map[PM_AfterGoal_Right] = GameSituation::gameSituation("goal_opp");

}

// Get player identifiers

int WorldModel::gamePace( ) const
{
    return M_game_pace;
}

ScoreAssessment WorldModel::scoreAssessment() const
{
    return M_score_assessment;
}

void WorldModel::updateScoreAssessment()
{
    M_score_assessment = ST_MEDIUM;
    int dif = 2;  //draw
    if ( M_relative_score > dif) M_score_assessment = ST_VERY_GOOD;  //
    else if ( M_relative_score > 0 ) M_score_assessment = ST_GOOD;
    else if ( M_relative_score < -dif ) M_score_assessment = ST_VERY_BAD;  //loose bad
    else if ( M_relative_score < 0 ) M_score_assessment = ST_BAD;

}

void WorldModel::updateGamePace( )
{
    int relative_score = relativeScore();
    bool attack = false;
    if ( M_world->ball().pos().x > 0.0)
       attack = true;
    bool gameEnd = false;
    if ( M_current_time > 0.9 * (ServerParam::i().actualHalfTime()*2))
       gameEnd = true;
    if ((relative_score == 0 || relative_score == -1) && gameEnd && attack)
       M_game_pace = 4;
    else if (relative_score == -1 && gameEnd)
       M_game_pace = 3;
    else if (relative_score == 0 && gameEnd && !attack)
       M_game_pace = 1;
    else if (relative_score == 1 && gameEnd)
       M_game_pace = 0;
    else if (relative_score == 1)
       M_game_pace = 1;
    else if (relative_score >= 3)
       M_game_pace = 3;
    else
       M_game_pace = 2;
}

/// PlayerToPointDistCmp

WorldModel::PlayerToPointDistCmp::PlayerToPointDistCmp( const AbstractPlayerCont &players,
                                                        Vector2D point)
    : M_players( players )
    , M_point( point )
{

}

void WorldModel::PlayerToPointDistCmp::setPoint(rcsc::Vector2D point)
{
    M_point = point;
}

void WorldModel::PlayerToPointDistCmp::updateCache()
{
    M_cached_distances.clear();
    for ( AbstractPlayerCont::const_iterator itp = M_players.begin();
          itp != M_players.end() ; itp++ )
    {
        const AbstractPlayerObject* player = *itp;
        M_cached_distances.insert(pair<unsigned int, double>( player->unum(),
                                                              player->pos().dist2( M_point )));
    }
}

    /*!
      \brief operation function
      \param lhs left hand side variable
      \param rhs right hand side variable
      \return compared result
     */
bool WorldModel::PlayerToPointDistCmp::operator()( unsigned int lhs,
                                                   unsigned int rhs ) const
{
    return  M_cached_distances.at(lhs) <= M_cached_distances.at(rhs);
}

/// Implementation of proxy method for Strategy World Model

/*::PlayerAffiliation WorldModel::ballOwner() const
{
    // TODO Define how the ball owner should be calculated
    // For now it will be the fastest player to intercept the ball

    int our_fastest_to_ball = 0, our_ball_reach_cycle = 0;

    if ( M_world_model->interceptTable()->fastestTeammate() == NULL)
    {
        our_fastest_to_ball = M_world_model->self().unum();
        our_ball_reach_cycle = M_world_model->interceptTable()->selfReachCycle();
    }
    else if ( M_world_model->interceptTable()->selfReachCycle() <= M_world_model->interceptTable()->teammateReachCycle() )
    {
        if ( M_world_model->interceptTable()->selfReachCycle() == M_world_model->interceptTable()->teammateReachCycle() )
        {
            // Player with lowest id is closer to the ball
            our_fastest_to_ball = M_world_model->self().unum() < M_world_model->interceptTable()->fastestTeammate()->unum() ?
                                  M_world_model->self().unum() :
                                  M_world_model->interceptTable()->teammateReachCycle();
        }
        else
        {
            our_fastest_to_ball = M_world_model->self().unum();
        }

        our_ball_reach_cycle = M_world_model->interceptTable()->selfReachCycle();
    }
    else
    {
        our_fastest_to_ball = M_world_model->interceptTable()->fastestTeammate()->unum();
        our_ball_reach_cycle = M_world_model->interceptTable()->teammateReachCycle();
    }

    if ( M_world_model->interceptTable()->fastestOpponent() != NULL &&
         our_ball_reach_cycle > M_world_model->interceptTable()->opponentReachCycle())
    {
        PlayerAffiliation pid_bowner( TA_OPP,
                             M_world_model->interceptTable()->fastestOpponent()->unum());
        dlog.addText( Logger::WORLD,
                     __FILE__ " Bowner (%s,%d) %d",
                     pid_bowner.team_affiliation() == TA_OUR ? "our" : "opp",
                     pid_bowner.number());
        return pid_bowner;
    }

    // Assume that self is closer to the ball otherwise
    dlog.addText( Logger::WORLD,
                  __FILE__ " Bowner (our,%d)",
                  M_world_model->self().unum());
    return PlayerAffiliation( TA_OUR, M_world_model->self().unum());
} */

}
