#ifndef WORLD_MODEL_H
#define WORLD_MODEL_H

#include <soccer/iworld_model.h>
#include <rcsc/player/world_model.h>

namespace soccer {

class WorldModel
        : public soccer::IWorldModel
{
public:
    WorldModel( const rcsc::WorldModel* world );

    void update();

    unsigned int currentTime() const;
    unsigned int remainingTimeInPeriod() const;

    unsigned int ballInterceptionTime( const TeamAffiliation& affiliation,
                                       const Unum& number ) const;
    unsigned int playerInterceptionTime( const TeamAffiliation& team1,
                                         const Unum& unum1,
                                         const TeamAffiliation& team2,
                                         const Unum& unum2) const;


    GameSituation gameSituation() const;

    // Game stopped?
    GameMode currentGameMode() const;

    /*!
      \brief Determine the current soccer moment of the game
      */
    SoccerMomentAffiliation soccerMoment() const;

    /*!
      \brief Determine the current time period
      */
    TimePeriod timePeriod() const;

    /// COMMON REGIONS
    rcsc::Rect2D field( ) const;
    double defenseLine(TeamAffiliation team_affiliation) const;
    double offenseLine(TeamAffiliation team_affiliation) const;
    double offsideLine(TeamAffiliation team_affiliation) const
    {
        return defenseLine( oppositeAffiliation( team_affiliation ) );
    }

    double goalLine(TeamAffiliation team_affiliation) const;
    double endLine(TeamAffiliation team_affiliation) const;

    rcsc::Rect2D goalieArea(TeamAffiliation team_affiliation) const;
    rcsc::Rect2D penaltyArea(TeamAffiliation team_affiliation) const;
    rcsc::Rect2D midfieldArea(TeamAffiliation team_affiliation) const;

    /// BALL INFORMATION
    rcsc::Vector2D ballPosition( ) const;
    double ballSpeed() const;
    rcsc::AngleDeg ballDirection() const;

    /// PLAYER INFORMATION
    Unum self() const;
    Unum goalie( TeamAffiliation team_affiliation ) const;
    rcsc::Vector2D playerPosition( const TeamAffiliation& affiliation, const Unum& number ) const;
    double playerSpeed( const TeamAffiliation& affiliation, const Unum& number ) const;
    rcsc::AngleDeg playerBodyDirection( const TeamAffiliation& affiliation, const Unum& number ) const;
    int playerTimeToReachPosition( const TeamAffiliation& affiliation,
                                           const Unum& number,
                                           const rcsc::Vector2D& position ) const;
    int playerTimeToReachMaxSpeed( const TeamAffiliation& affiliation,
                                           const Unum& number ) const;
    double playerMaxSpeed( const TeamAffiliation& affiliation,
                                   const Unum& number ) const;


    std::vector<Unum> activePlayers( TeamAffiliation team_affiliation ) const;

    PlayerAffiliation playerFastestTo( const rcsc::Vector2D& position );
    PlayerAffiliation playerFastestTo( const rcsc::Region2D& region );
    PlayerAffiliation playerFastestTo( const TeamAffiliation &team, const Unum &unum );
    PlayerAffiliation playerFastestToBall( );

    Unum teammateFastestTo( const rcsc::Vector2D& position );
    Unum teammateFastestTo( const rcsc::Region2D& region );
    Unum teammateFastestTo( const TeamAffiliation& team,
                            const Unum& unum );
    Unum teammateFastestToBall( );

    Unum opponentFastestTo( const rcsc::Vector2D& position );
    Unum opponentFastestTo( const rcsc::Region2D& region );
    Unum opponentFastestTo( const TeamAffiliation& team,
                            const Unum& unum );
    Unum opponentFastestToBall( );

    std::vector<Unum> teammatesFromPlayer( const TeamAffiliation& team_affiliation,
                                           const Unum& number) const;
    std::vector<Unum> opponentsFromPlayer( const TeamAffiliation& team_affiliation,
                                           const Unum& number ) const;

    std::vector<Unum> playersFromBall( TeamAffiliation team_affiliation ) const;
    std::vector<Unum> playersFromGoal( TeamAffiliation team_affiliation,
                                       TeamAffiliation goal_affiliation) const;
    std::vector<Unum> playersBehindBall( TeamAffiliation team_affiliation) const;
    std::vector<Unum> playersInFrontOfBall( TeamAffiliation team_affiliation) const;

    // std::vector<IPlayerId*> teammatesToPlayer( const IPlayerId& player_id ) const;
    // std::vector<IPlayerId*> teammatesToPosition( const rcsc::Vector2D& position ) const;
    PlayerAffiliation lastKicker() const;

    TeamAffiliation teamInPossession() const;


    unsigned int goalsScored(TeamAffiliation affiliation) const;

    int relativeScore() const
    {
        return goalsScored( TA_OUR ) - goalsScored( TA_OPP );
    }

    /*!
      \brief Determine the players belonging to a particular team affiliation that are offside
      \param team_affiliation The team affiliation the players belong to
      \returns Players that are offside
      */
    std::vector<Unum> playersOffside(TeamAffiliation team_affiliation) const;
    /*!
      \brief Determine the players belonging to a particular team affiliation that are not offside
      \param team_affiliation The team affiliation the players belong to
      \returns Players that are not offside
      */
    std::vector<Unum> playersNotOffside(TeamAffiliation team_affiliation) const;


    int gamePace( ) const;

    ScoreAssessment scoreAssessment() const;
protected:
      // Helper method for update
      void updatePlayersRelativeToBall( );
      void upatePlayerPositions( );
      // void updateGameSituation( );
      void updateScoreAssessment( );
      void updateGameMode( );
      void updatePlayersCalculus( );
      void updateGamePace( );
      void updateScore( );
      void updateSoccerMoment();

      /// Updater functions
      void setSoccerMoment( const SoccerMomentAffiliation& soccer_moment );

      const rcsc::AbstractPlayerCont& playersBehindBall( rcsc::SideID team_side);
      const rcsc::AbstractPlayerCont& oursBehindBall( );
      const rcsc::AbstractPlayerCont& theirsBehindBall( );


      void initPlayModes();
private:
      /// Override IWorldModel
      rcsc::SideID side( const TeamAffiliation &affiliation) const;
      const rcsc::AbstractPlayerObject* getAbstractPlayerObject( const PlayerAffiliation& player ) const;

      void updatePlayerOffside( std::vector<unsigned int>& players_offside,
                                std::vector<unsigned int>& players_not_offside,
                                double offside_line,
                                const rcsc::AbstractPlayerObject& getAbstractPlayerObject  );
      void updatePlayerRelativeToBallX( std::vector<unsigned int>& players_behind_ball,
                                         std::vector<unsigned int>& players_in_front_ball,
                                         const rcsc::AbstractPlayerObject& player  );
      void updatePlayersFromBall( std::vector<unsigned int>& player_ids_from_ball,
                                  const rcsc::PlayerPtrCont& players_from_ball);

      void updatePlayers( const rcsc::AbstractPlayerObject* player );


      /*!
        \class PlayerToPointDistCmp
        \brief predicate functor to compare player's distance to their goal
      */
      class PlayerToPointDistCmp
          : public std::binary_function< const unsigned int,
                                         const unsigned int,
                                         bool > {
      public:
          PlayerToPointDistCmp( const rcsc::AbstractPlayerCont& activePlayers,
                                rcsc::Vector2D point);

          /*!
            \brief operation function
            \param lhs left hand side variable
            \param rhs right hand side variable
            \return compared result
           */
          bool operator()( unsigned int lhs,
                           unsigned int rhs ) const;

          void updateCache();
          void setPoint(rcsc::Vector2D point);
      private:
          const rcsc::AbstractPlayerCont& M_players;
          rcsc::Vector2D M_point;
          std::map<unsigned int, double> M_cached_distances;
      };

      unsigned int M_current_time;
      unsigned int M_remaining_time;
      TimePeriod M_time_period;
      SoccerMomentAffiliation M_soccer_moment;


      const rcsc::WorldModel* M_world;

      /// COMMON REGIONS
      /// Static regions
      rcsc::Rect2D M_field;
      rcsc::Rect2D M_left_goalie_area;
      rcsc::Rect2D M_right_goalie_area;
      rcsc::Rect2D M_left_penalty_area;
      rcsc::Rect2D M_right_penalty_area;
      rcsc::Rect2D M_left_midfield_area;
      rcsc::Rect2D M_right_midfield_area;

      double M_left_end_line;
      double M_right_end_line;
      double M_left_goal_line;
      double M_right_goal_line;

      /// Variable lines/regions
      double M_our_defense_line;
      double M_opp_defense_line;
      double M_our_offense_line;
      double M_opp_offense_line;


      /// Players calculations => Sorted by distances
      std::vector<unsigned int> M_our_players_from_ball;
      std::vector<unsigned int> M_opp_players_from_ball;

      std::vector<unsigned int> M_our_players_from_own_goal;
      std::vector<unsigned int> M_opp_players_from_own_goal;

      std::vector<unsigned int> M_our_players_to_opp_goal;
      std::vector<unsigned int> M_opp_players_to_our_goal;

      std::vector<unsigned int> M_our_players_behind_ball;
      std::vector<unsigned int> M_opp_players_behind_ball;

      std::vector<unsigned int> M_our_players_in_front_of_ball;
      std::vector<unsigned int> M_opp_players_in_front_of_ball;

      std::vector<unsigned int> M_our_players_offside;
      std::vector<unsigned int> M_opp_players_offside;

      std::vector<unsigned int> M_our_players_non_offside;
      std::vector<unsigned int> M_opp_players_non_offside;

      GameMode M_previous_game_mode;
      GameMode M_current_game_mode;

      unsigned int M_our_score;
      unsigned int M_opp_score;
      int M_relative_score;
      int M_game_pace;
      ScoreAssessment M_score_assessment;

      PlayerToPointDistCmp M_teammates_from_own_goal_cmp;
      PlayerToPointDistCmp M_opponents_from_own_goal_cmp;

      //
      typedef std::map<rcsc::PlayMode, GameSituation> PlayModeToGameSituationMap;
      PlayModeToGameSituationMap M_play_modes_map;
};

}

#endif // WORLD_MODEL_H
