// -*-c++-*-
#ifndef SP_IMESSAGE_H
#define SP_IMESSAGE_H

#include <cassert>
#include <memory>
#include <iostream>

#include <setplay/communication/imessage_processor.h>
#include <setplay/model/participants/player_affiliation.h>

#include <setplay/generators/sexpr/setplay_generator_sexpr.h>

namespace setplay {

    typedef enum {
        MSG_SETPLAY_START,
        MSG_STEP_ENTER,
        MSG_STEP_TRANSITION,
        MSG_ACTION_DONE,
        MSG_ABORT,
        MSG_FINISH
    } MessageType;

    class IMessage {
    public:
      typedef std::map<std::string, Object*> Arguments;

      IMessage(const PlayerAffiliation& sender, unsigned int time, unsigned int setplay_id, bool inverted)
          : M_sender( sender )
          , M_time( time )
          , M_setplay_id( setplay_id )
          , M_inverted( inverted )
      {

      }

      IMessage(const IMessage& other)
          : M_sender( other.M_sender )
          , M_time ( other.M_time )
          , M_setplay_id( other.M_setplay_id )
          , M_inverted( other.M_inverted )
      {
          if ( !other.M_arguments.empty() )
          {
              for ( map<string, Object*>::const_iterator it = other.M_arguments.begin();
                    it != other.M_arguments.end();
                    it++)
              {
                  M_arguments.insert( pair<string, Object*>( it->first,
                                                             it->second->clone() ));
              }
          }
      }

      virtual ~IMessage()
      {
          if ( !M_arguments.empty() )
          {
              // Delete each contained object
              for ( Arguments::iterator it = M_arguments.begin();
                    it != M_arguments.end();
                    it++)
              {
                  delete it->second;
              }
          }
      }

      const PlayerAffiliation& sender() const
      {
          return M_sender;
      }
	  
      unsigned int setplayId() const
	  {
        return M_setplay_id;
	  }

      bool inverted() const
      {
          return M_inverted;
      }

      unsigned int time() const
      {
          return M_time;
      }

      bool numArguments() const
      {
          return M_arguments.size();
      }

      const Object* argument( const std::string& name ) const
      {
          Arguments::const_iterator it = M_arguments.find( name );

          assert ( it != M_arguments.end() );

          return it->second;
      }

      const Arguments& arguments() const
      {
          return M_arguments;
      }

      void push(std::string name, Object* object)
      {
          M_arguments.insert(pair<string, Object*>(name, object));
      }

      void setSetplayId( unsigned setplay_id )
      {
          M_setplay_id = setplay_id;
      }

      void setInverted( bool inverted )
      {
          M_inverted = inverted;
      }

      virtual void print(std::ostream& os) const
      {
          os << "Player: ("
             << M_sender.affiliation() << "," << M_sender.number() << ")"
             << " Time: " << M_time
             << " Inverted: " << (M_inverted ? "true" : "false");
          os << " Arguments: ";
          if ( M_arguments.empty() )
          {
              os << "none";
          }
          else
          {
              SetplayGeneratorSexpr sexpr( os );
              for ( Arguments::const_iterator ita = M_arguments.begin();
                    ita != M_arguments.end() ; ita++ )
              {
                  os << ita->first;
                  os << "=";
                  ita->second->visit( sexpr );
                  os << ", ";
              }
          }
      }

      virtual void getProcessed(IMessageProcessor& processor) const = 0;

      virtual MessageType type() const = 0;
      virtual IMessage* clone() const = 0;

    private:
        PlayerAffiliation M_sender; /// Each message is sent by a specific participant
        unsigned int M_time; /// The time at which the message is sent/received
        int M_setplay_id; /// All messages must refer to a plan id enable disambiguation
        bool M_inverted; /// Is the set-play inverted?
        Arguments M_arguments; /// Run-time arguments of the set-play
    };

}

#endif
