// -*-c++-*-
#include "cbr.h"

#include <fstream>

#include "iworld_model.h"
#include "geom/IPoint2D.h"
#include "icondition.h"

#include <boost/spirit/include/classic_core.hpp>
#include <boost/spirit/include/classic_assign_actor.hpp>

#include "parsers/simple_parsers.h"
#include <setplay/runtime/setplay_interpreter.h>

#define EVAL_WHEN_NO_TRIES 100

using namespace boost;
using namespace BOOST_SPIRIT_CLASSIC_NS;

namespace setplay {

    CaseEvaluator::RegionPositions S_transversal_regions = createTransversalRegions();
    CaseEvaluator::RegionPositions S_longitudinal_regions = createLongitudinalRegions();
    CaseEvaluator::RegionPositions S_radial_regions = createRadialRegions();

    CaseEvaluator::RegionPositions CaseEvaluator::createTransversalRegions()
    {
        CaseEvaluator::RegionPositions regions;

        regions.insert(pair<std::string, unsigned int>("our_back", 0));
        regions.insert(pair<std::string, unsigned int>("our_middle", 1));
        regions.insert(pair<std::string, unsigned int>("our_front", 2));
        regions.insert(pair<std::string, unsigned int>("their_front", 3));
        regions.insert(pair<std::string, unsigned int>("their_middle", 4));
        regions.insert(pair<std::string, unsigned int>("their_back", 5));

        return regions;
    }

    CaseEvaluator::RegionPositions CaseEvaluator::createLongitudionalRegions()
    {
        CaseEvaluator::RegionPositions regions;

        regions.insert(pair<std::string, unsigned int>("far_left", 0));
        regions.insert(pair<std::string, unsigned int>("mid_left", 1));
        regions.insert(pair<std::string, unsigned int>("centre_left", 2));
        regions.insert(pair<std::string, unsigned int>("centre_right", 3));
        regions.insert(pair<std::string, unsigned int>("mid_right", 4));
        regions.insert(pair<std::string, unsigned int>("far_right", 5));

        return regions;
    }

    CaseEvaluator::RegionPositions CaseEvaluator::createRadialRegions()
    {
        CaseEvaluator::RegionPositions regions;

        regions.insert(pair<std::string, unsigned int>("sl_1", 0));
        regions.insert(pair<std::string, unsigned int>("sl_2", 1));
        regions.insert(pair<std::string, unsigned int>("sl_3", 2));
        regions.insert(pair<std::string, unsigned int>("sl_4", 3));
        regions.insert(pair<std::string, unsigned int>("sl_5", 4));
        regions.insert(pair<std::string, unsigned int>("sl_6", 5));
        regions.insert(pair<std::string, unsigned int>("sl_7", 6));
        regions.insert(pair<std::string, unsigned int>("sl_8", 7));
        regions.insert(pair<std::string, unsigned int>("sl_9", 8));
        regions.insert(pair<std::string, unsigned int>("sr_1", 9));
        regions.insert(pair<std::string, unsigned int>("sr_2", 10));
        regions.insert(pair<std::string, unsigned int>("sr_3", 11));
        regions.insert(pair<std::string, unsigned int>("sr_4", 12));
        regions.insert(pair<std::string, unsigned int>("sr_5", 13));
        regions.insert(pair<std::string, unsigned int>("sr_6", 14));
        regions.insert(pair<std::string, unsigned int>("sr_7", 15));
        regions.insert(pair<std::string, unsigned int>("sr_8", 16));
        regions.insert(pair<std::string, unsigned int>("sr_9", 17));

        return regions;
    }

    float CaseEvaluator::weight(const Case &the_case)
    {
        float weight = 0.0;
        // When kick-off, location of ball is irrelevant, therefore areas will
        // not be compared...
        if ( M_game_mode== KickOff_Our     || M_game_mode== KickOff_Opp ||
             M_game_mode== GoalieCatch_Our || M_game_mode== GoalieCatch_Opp )
        {
            weight = 1.0;
        }
        else
        {
            // TODO This should not rely on the world!!
            const string& transversal_region = getTransversalRegionName( *world.ball_position() );
            const string& longitudinal_region= getLongitudinalRegionName( *world.ball_position() );
            const string& radial_region = getRadialRegionName( *world.ball_position() );


            // weight should have a maximum of 1 and minimum of 0, and be inversely
            // proportional to the square of the distance. Therefore: 0.5^d
            return pow(0.5,
                       (int) longitudinalRegionsDistance(longitudinal_region, the_case.longitudinalRegion()) +
                             transversalRegionsDistance(transversal_region, the_case.transversalRegion()) +
                             radialRegionsDistance(radial_region, the_case.radialRegion()));
        }

        //DEBUG
        cerr << "SP: CBR weight: " << weight << endl;
        return weight;
    }

    const std::string& CaseEvaluator::getLongitudinalRegionName(const rcsc::Vector2D& position) const
    {
        getRgetRegionName( S_longitudinal_regions, position );
    }

    const std::string& CaseEvaluator::getTransversalRegionName(const rcsc::Vector2D& position) const
    {
        getRegionName( S_transversal_regions, position );
    }

    const std::string& CaseEvaluator::getRadialRegionName(const rcsc::Vector2D& position) const
    {
        getRegionName( S_radial_regions, position );
    }

    const std::string& CaseEvaluator::getRegionName( const RegionPositions& regions,
                                                     const rcsc::Vector2D& position) const
    {
        for ( RegionPositions::const_iterator it = regions.begin();
              it != regions.end(); it++ )
        {
            VarRegion var_region( it->first );
            rcsc::Region2D* region = M_interpreter->interpret( var_region );
            if ( region->contains( position ) )
            {
                delete region;
                return it->first;
            }
            delete region;
        }
        asser(false);
        return "unknown";
    }

    float CaseEvaluator::evaluate(const Case& the_case)
    {

    }


    /*Case* Case::parse(const string& in, string& out){
      string name,regTrans,regLong,regRad,mode,adversary;
      int numTries,numFail,numTriesThisGame,numFailThisGame;

      // DEBUG
      //cerr<<"SP: CBR in Case constructor: "<<in<<endl;



      rule<> case_p
        = *space_p >> str_p("(case") >> *space_p >> ":setplayName" >> *space_p
                   >> identifier_p[assign_a(name)] >> *space_p >> ":transverse"
                   >> *space_p >> identifier_p[assign_a(regTrans)]
                   >> *space_p >> ":longitudinal"
                   >> *space_p >> identifier_p[assign_a(regLong)]
                   >> *space_p >> ":radial"
                   >> *space_p >> identifier_p[assign_a(regRad)]
                   >> *space_p>> ":gameMode"
                   >> *space_p>> identifier_p[assign_a(mode)]
                   >> *space_p>> ":adversary"
                   >> *space_p>> identifier_p[assign_a(adversary)]
                   >> *space_p>> ":numTries">>*space_p>>int_p[assign_a(numTries)]
                   >> *space_p>> ":numFail">>*space_p>>int_p[assign_a(numFail)]
                   >> *space_p>> ":numTriesThisGame"
                   >> *space_p>> int_p[assign_a(numTriesThisGame)]
                   >> *space_p>> ":numFailThisGame"
                   >> *space_p>> int_p[assign_a(numFailThisGame)]
                   >> *space_p>> ")";

      parse_info<> result = BOOST_SPIRIT_CLASSIC_NS::parse(in.c_str(),case_p,
                                                           nothing_p);


       PlayMode* gm = CondPlayMode::parsePlayMode(mode);

      if(result.hit && gm!=NULL){
        out=in.substr(result.length);
        return new Case(name,regTrans,regLong,regRad,
                        *gm,adversary,
                        numTries+numTriesThisGame,
                        // Sum the results in the last game to the total
                        numFail+numFailThisGame,0,0);
      }

      cerr<<"SP: Bad case parse!"<<endl;
      out=in;
      return NULL;
    }*/

    const Case::Evaluation Case::evaluate(const IContextProvider& world) const
    {
      double weight = 0;

      // When kick-off, location of ball is irrelevant, therefore areas will
      // not be compared...
      if ( M_game_mode== KickOff_Our     || M_game_mode== KickOff_Opp ||
           M_game_mode== GoalieCatch_Our || M_game_mode== GoalieCatch_Opp )
      {
          weight = 1;
      }
      else
      {
          const string transverseCase = world.getTransversalRegion( *world.ball_position() );
          const string longitudinalCase= world.getLongitudinalRegion( *world.ball_position() );
          const string radialCase = world.getRadialRegion( *world.ball_position() );


          // weight should have a maximum of 1 and minimum of 0, and be inversely
          // proportional to the square of the distance. Therefore: 0.5^d
          weight = pow(0.5,
                   (int)(world.longitudinalRegionDistance(longitudinalCase, M_longitudinal_region) +
                         world.transverseRegionDistance(transverseCase, M_transversal_region) +
                         world.radialRegionDistance(radialCase, M_radial_region)));

          //DEBUG
          cerr << "SP: CBR weights:"
               << world.longitudinalRegionDistance(longitudinalCase, M_longitudinal_region)
               << ","<< world.transverseRegionDistance(transverseCase, M_transversal_region)
               << ","<< world.radialRegionDistance(radialCase, M_radial_region)
               << " total:"<<weight<<endl;
      }

      // Determine where this case should be counted: if it is against this team,
      // then both in this game and this team. Else, just in other teams
      // General rule for evaluation: when no tries are recorded in this game, 0 will
      // be awarded. When no fails, then the Setplay has been
      // highly successful, therefore should get a high score. The standard calculation
      // cannot be used, since it would be a division by 0. Therefore, a "0.5 imaginary
      // fail" will be considered, which will make the score go high as it is tried with no
      // urther fails. Else, the ratio total/fails will be
      // awarded, to ensure that cases with all fails will have a low value (1),
      // which can still be selected, in order to ensure (infrequent) retrial of these
      // cases.
      Evaluation eval;
      if ( world.opponent_name() == M_adversary )
      {
          if ( M_num_tries_this_game == 0 )
            eval.evalThisGame=0;
          else if(M_num_fails_this_game==0)
            eval.evalThisGame=(double)M_num_tries_this_game/0.5;
          else
            eval.evalThisGame=(double)M_num_tries_this_game/M_num_fails_this_game;

          eval.weightThisGame=weight;

          if(M_num_tries==0)
            eval.evalThisTeam=0;
          else if(M_num_failures==0)
            eval.evalThisTeam=(double)M_num_tries/0.5;
          else
            eval.evalThisTeam=(double)M_num_tries/M_num_failures;

          eval.weightThisTeam=weight;

          eval.evalOtherTeams=0;
          eval.weightOtherTeams=0;
      }
      else
      {
          eval.evalThisGame=0;
          eval.weightThisGame=0;
          eval.evalThisTeam=0;
          eval.weightThisTeam=0;
          if(M_num_tries==0)
              eval.evalOtherTeams=0;
          else if(M_num_failures==0)
              eval.evalThisTeam=(double)M_num_tries/0.5;
          else
              eval.evalThisTeam=(double)M_num_tries/M_num_failures;
          eval.weightOtherTeams=weight;
      }

      return eval;
    }


    double Case::similarity(const Case& other) const
    {
      //cerr<<"SB: CBR game mode comparison, "<<gameMode<<"=="<<KickOff_Our<<":"
      //  <<gameMode==KickOff_Our<<endl;
      // When kick-off or goalie catch, location of ball is irrelevant, therefore
      // areas will not be compared...
      if ( M_id == other.M_id && M_adversary == other.M_adversary &&
             M_game_mode == other.M_game_mode &&
             ( M_game_mode == KickOff_Our || M_game_mode == KickOff_Opp ||
               M_game_mode == GoalieCatch_Our || M_game_mode == GoalieCatch_Opp ||
             ( M_transversal_region == other.M_transversal_region &&
               M_longitudinal_region == other.M_longitudinal_region &&
               M_radial_region == other.M_radial_region)) )
      {
          return 1;
      }
      return 0;
    }

    /*void Case::write(ostream &os) const{
      os <<"(case :setplayName "<< M_name <<" :transverse "<<transverseRegionName
         <<" :longitudinal "<<longitudinalRegionName<<" :radial "<<radialRegionName
         <<" :gameMode " <<gameMode <<" :adversary "<<adversary
         <<" :numTries " <<numTries<<" :numFail "<<numFail
         <<" :numTriesThisGame "<<numTriesThisGame<<" :numFailThisGame "
         <<numFailThisGame<<") ";
    }*/

    std::ostream& operator<<( std::ostream & os, const Case::Evaluation& e )
    {
        return os << "(eval :thisGame " << e.evalThisGame
                  << "/" << e.weightThisGame
                  << " :thisTeam " << e.evalThisTeam
                  << "/" << e.weightThisTeam
                  << " :otherTeams " << e.evalOtherTeams
                  << "/" << e.weightOtherTeams
                  << ")";
    }


    CaseBase::CaseBase(const std::string& ourTeam_,const string& fileName_,
                       bool writeToFile_, bool writeToFileAfterExecution_,
                       const double weightThisGame_,
                       const double weightThisTeam_,
                       const double weightOtherTeams_)
      : ourTeam(ourTeam_)
      , fileName(fileName_)
      , writeCasesToFile(writeToFile_)
      , writeToFileAfterExecution(writeToFileAfterExecution_)
      , weightThisGame(weightThisGame_)
      , weightThisTeam(weightThisTeam_)
      , weightOtherTeams(weightOtherTeams_)
    {
      ///LMOTA: read conf file
      /*vector<string> text;
      string line;
      ifstream textstream(fileName.c_str());

      while (getline(textstream, line)) {
        text.push_back(line);
      }
      textstream.close();
      string alltext;
      for (unsigned int i=0; i < text.size(); i++){
        // substitute tabs by spaces
        size_t j = text[i].find('\t', 0);
        while(j!=string::npos){
          text[i].replace(j,1," ");
          j = text[i].find("\t", 0);
        }
        alltext += text[i];
      }

      // DEBUG
      //cerr<<"SP: CBR case base being read:"<<alltext<<endl;

      // Check if nothing was read...
      if(alltext.length()==0)
        cerr<<"SP: empty CaseBase file!"<<endl;
      else{
        string rest;

        fcportugal::setplay::Case* caseTemp=Case::parse(alltext,rest);

        while(caseTemp){
          caseBase[caseTemp->M_name()].push_back(caseTemp);

          caseTemp=Case::parse(rest,rest);
        }

        //DEBUG
        if(rest.length()>0)
          cerr<<"\n Remainder of CaseBase Config file:"<<rest<<endl;
      }*/
    }

    double
    CaseBase::evalSetplays(vector<CaseBase::SetplayID>& possibleSetplays,
                           const IContextProvider& world) const{
      //DEBUG
      cerr<<"SP: evaluating "<<possibleSetplays.size()<<" setplays"<<endl;
      double sumEvaluations=0;

      for(unsigned int i=0;i!=possibleSetplays.size();i++)
      {
        double totalEval=0;
        map<std::string,std::vector<Case*> >::const_iterator it
          = M_setplay_cases.find(possibleSetplays.at(i).name);
        if( it==M_setplay_cases.end() ) // Award standard value for Setplays with no registered executions
          totalEval=EVAL_WHEN_NO_TRIES;
        else
        {
          double thisGameEvalSum=0,
                  thisGameWeightSum=0,
                  thisTeamEvalSum=0,
                  thisTeamWeightSum=0,
                  otherTeamsEvalSum=0,
                  otherTeamsWeightSum=0;
          for(unsigned int j=0;j!=it->second.size();j++)
          {
              Evaluation thisEval = it->second.at(j)->evaluate( world );

            //DEBUG
            cerr << "SP: CBR case " << possibleSetplays.at(i).name
                 << " eval:" << thisEval
                 << endl;

            thisGameEvalSum += thisEval.evalThisGame*thisEval.weightThisGame;
            thisGameWeightSum += thisEval.weightThisGame;
            thisTeamEvalSum+=thisEval.evalThisTeam*thisEval.weightThisTeam;
            thisTeamWeightSum+=thisEval.weightThisTeam;
            otherTeamsEvalSum+=thisEval.evalOtherTeams*thisEval.weightOtherTeams;
            otherTeamsWeightSum+=thisEval.weightOtherTeams;
          }

          totalEval
            // In this game, the eval can be 0, since the case may refer to tries
            // with other opponents
            = (thisGameWeightSum == 0 ? 0 : thisGameEvalSum / thisGameWeightSum )
            * weightThisGame
            + (thisTeamWeightSum == 0 ? 0 : thisTeamEvalSum / thisTeamWeightSum )
            * weightThisTeam
            + (otherTeamsWeightSum == 0 ? 0 : otherTeamsEvalSum / otherTeamsWeightSum)
            * weightOtherTeams;
        }
        possibleSetplays.at(i).eval=totalEval;
        sumEvaluations += totalEval;
      }
      return sumEvaluations;

    }

    unsigned int
    CaseBase::selectSetplay(vector<SetplayID>& possibleSetplayNums,
                            const IContextProvider& world) const
    {
      // computes and puts evaluation in each of the vector's members
      double sumEvaluations = evalSetplays(possibleSetplayNums,world);

      //DEBUG
      //cerr<<"SP: sum of evaluations:"<< sumEvaluations<<endl;

      // Choose the setplay by a random number between 0 and sumEvaluations:
      // this way the evaluation scores will be the relative probability of
      // a setplay being chosen
      double randDouble = rand() * sumEvaluations / RAND_MAX;

      //DEBUG
      cerr << "SP: CBR choosing SP, finding "
           << randDouble << " in "
           << sumEvaluations
           << " size:" << possibleSetplayNums.size()
           << endl;
      double evaluationAccum=0;
      for(unsigned int i=0; i!=possibleSetplayNums.size();i++)
        if((evaluationAccum+=possibleSetplayNums[i].eval)>=randDouble)
          return possibleSetplayNums[i].id;


      cerr << "SP: CBR Bad Setplay chosen by CBR!" << endl;
        return 9999;
    }


    void CaseBase::registerCase( const unsigned int& setplay_id,
                                 const rcsc::Vector2D ball_position,
                                 const PlayMode& gm,
                                 bool success,
                                 const IContextProvider& world){
      //DEBUG
      //cerr<<"CBR: size of case base before registering:"<<caseBase.size()<<endl;

      // if ball position is relevant, determine it. Irrelevant for kick-offs and
      // goalie catchs...
      bool irrelevant = ( gm == KickOff_Our ||
                          gm == KickOff_Opp ||
                          gm == GoalieCatch_Our ||
                          gm == GoalieCatch_Opp);

      if ( irrelevant )
      {
          return;
      }

        IPoint2D* normalizedPoint= ball_position->clone(); // HACK This should only return the point
        // Normalize the point if it is outside the field, for correct classification
        if(!irrelevant)
        {
            // TODO This is code is far from optimized => REFACTOR THIS AND FIX LOGIC
            if( normalizedPoint->getX()->value()>world.field_length()/2 )
              normalizedPoint->setVec(new Decimal(world.field_length()/2), normalizedPoint->getY());
            if(normalizedPoint->getX()->value()<-world.field_length()/2)
                normalizedPoint->setVec(new Decimal(-world.field_length()/2), normalizedPoint->getY());
            if(normalizedPoint->getY()->value()>world.field_width()/2)
              normalizedPoint->setVec(normalizedPoint->getX(), new Decimal(world.field_width()/2));
            if(normalizedPoint->getY()->value()<-world.field_width()/2)
              normalizedPoint->setVec(normalizedPoint->getX(), new Decimal(-world.field_width()/2));
        }


      const string  transverse
        = (irrelevant?"irrelevant":world.getTransversalRegion(normalizedPoint));
      const string longitudinal
        = (irrelevant?"irrelevant":world.getLongitudinalRegion(normalizedPoint));
      const string radial
        = (irrelevant?"irrelevant":world.getRadialRegion(normalizedPoint));

      // Check if region names were not found...
      if(transverse==""||longitudinal==""|| radial==""){
        cerr<<"CBR: ignoring case that could not be localised!"<<endl;
        return;
      }

      Case* new_case = new Case( setplayName,
                                 transverse,
                                 longitudinal,
                                 radial,
                                 gm,
                                 world.opponentName( ),
                                 0u, 0u, 1u, (success?0:1));

      bool insertion = false;

      // Find existing set-play cases
      SetplayCases::iterator itsc = M_setplay_cases.find( setplay_id );
      Cases* cases;
      if ( itsc == M_setplay_cases.end() )
      {
          cases = new Cases();
          cases->push_back( new_case );
          M_setplay_cases.insert(SetplayCases::value_type( setplay_id, cases ) );
          return;
      }

      cases = itsc->second;

      // Find the most similar case
      Case* most_similar_case = NULL;
      double best_similarity = 0.0;
      for ( SetplayCases::iterator it = cases->begin() ;
            it != cases->end(); it++)
      {
          double cases_similarity = new_case->similarity( *it );
          if ( cases_similarity > best_similarity )
          {
              most_similar_case = *it;
              best_similarity = cases_similarity;
          }
      }

      if ( best_similarity > 0.75 )
      {
          if ( success )
              most_similar_case->addSuccess( );
          else
              most_similar_case->addFailure( );
      }
      else
      {
          cases->push_back( new_case );
      }

      //DEBUG
      //cerr<<"CBR: size of case base after registering:"<<caseBase.size()<<endl;
      if(writeCasesToFile && writeToFileAfterExecution)
        writeToFile();

      //DEBUG
      //cerr<<"CBR: size of case base after writing:"<<caseBase.size()<<endl;
    }

    void CaseBase::write(std::ostream &os) const
    {
        for(std::map<std::string,std::vector<Case*> >::const_iterator it
            = M_setplay_cases.begin(); it != M_setplay_cases.end(); ++it)
        for( unsigned int i=0 ; i!=it->second.size() ; i++)
        {
            it->second.at(i)->write(os);
            os<<endl;
        }
    }

    void CaseBase::writeToFile() const
    {
        ofstream textstream(fileName.c_str());
        write(textstream);
    }

}
