// -*-c++-*-
#ifndef SP_CBR_H
#define SP_CBR_H

#include <map>
#include <vector>
#include <string>
#include <iostream>

#include <rcsc/geom/vector_2d.h>
#include <setplay/runtime/util.h>

using namespace std;

namespace setplay {
    class IWorldModel;
    class IPoint2D;
    class SetplayInterpreter;


    template<class C>
    class ICaseEvaluator
    {
    public:
        ~ICaseEvaluator()
        {

        }

        virtual float evaluate(const C& the_case) = 0;
    };

    /// Forward declare the case type
    class Case;
    class CaseEvaluator
            : public ICaseEvaluator<Case>
    {
    public:
        CaseEvaluator();

        float evaluate(const Case& the_case);
    private:
        typedef std::map<std::string, unsigned int> RegionPositions;

        static RegionPositions createTransverseRegions();
        static RegionPositions createLongitudionalRegions();
        static RegionPositions createRadialRegions();

        unsigned int longitudinalRegionsDistance(const std::string& region1,
                                                 const std::string& region2);
        unsigned int transversalRegionsDistance(const std::string& region1,
                                                const std::string& region2);
        unsigned int radialRegionsDistance(const std::string& region1,
                                           const std::string& region2);
        unsigned int regionsDistance( RegionPositions& regions,
                                      const std::string& region1,
                                      const std::string& region2,
                                      double exponent);

        float weight(const Case& the_case);
        const std::string& getLongitudinalRegionName(const rcsc::Vector2D& position) const;
        const std::string& getTransversalRegionName(const rcsc::Vector2D& position) const;
        const std::string& getRadialRegionName(const rcsc::Vector2D& position) const;

        const std::string& getRegionName( const RegionPositions& regions,
                                          const rcsc::Vector2D& position);

        // Member variables
        static const RegionPositions S_transversal_regions;
        static const RegionPositions S_longitudinal_regions;
        static const RegionPositions S_radial_regions;

        SetplayInterpreter* M_interpreter;
    };


    struct Evaluation
    {
        double evalThisGame;
        double weightThisGame;
        double evalThisTeam;
        double weightThisTeam;
        double evalOtherTeams;
        double weightOtherTeams;

        Evaluation()
            : evalThisGame(0)
            , weightThisGame(0)
            , evalThisTeam(0)
            , weightThisTeam(0)
            , evalOtherTeams(0)
            , weightOtherTeams(0)
        {

        }
    };

    class Case {
        public:

        Case( const unsigned int& id,
              const std::string& transversal_region,
              const std::string& longitudinal_region,
              const std::string& radial_region,
              const PlayMode game_mode,
              const std::string& adversary,
              const unsigned int num_tries = 0u,
              const unsigned int num_failures = 0u,
              const unsigned int numTriesThisGame_ = 0u,
              const unsigned int numFailThisGame_ = 0u)
        : M_id(id)
        , M_transversal_region(transversal_region)
        , M_longitudinal_region(longitudinal_region)
        , M_radial_region(radial_region)
        , M_game_mode(game_mode)
        , M_adversary(adversary)
        , M_num_tries(num_tries)
        , M_num_failures(num_failures)
        , M_num_tries_this_game(numTriesThisGame_)
        , M_num_fails_this_game(numFailThisGame_)
        {

        }

        ~Case()
        {

        }

        //Parser
        // static Case* parse(const std::string& in, std::string& out);

        // Output
        // void write(std::ostream &os) const;

        // Inspectors
        const unsigned int setplayId() const
        {
            return M_id;
        }

        // Comparison
        double similarity(const Case& otherCase) const;

        const Evaluation evaluate(const IWorldModel& world) const;

        // Modifiers
        void addSuccess()
        {
            M_num_tries_this_game++;
        }

        void addFailure()
        {
            M_num_tries_this_game++;
            M_num_fails_this_game++;
        }

        float success()
        {
            return 1 - failure();
        }

        float failure()
        {
            return M_num_failures * 1.0 / M_num_tries;
        }

        private:
            //Attributes
            unsigned M_id;

            // IRegion names
            std::string M_transversal_region;
            std::string M_longitudinal_region;
            std::string M_radial_region;

            PlayMode M_game_mode;
            std::string M_adversary;
            unsigned int M_num_tries;
            unsigned int M_num_failures;
            // The bottom two should not be in the the case!!!
            // A game case base should also exist and possibly merged with the existing one in the end
            unsigned int M_num_tries_this_game;
            unsigned int M_num_fails_this_game;
    };

    class CaseBase
    {
    public:
        // Arguments:
        // ourTeam - our team name
        // fileName - file where cases should be read and written
        // writeToFileAfterExecution - if set, write to file after each
        //   execution of a setplay
        // weightThisGame, weightThisTeam, weightOtherTeams -
        //   wheights given to cases in this game, against this same team,
        //   or against other teams
        // scoreIfZero - if a case has an evaluation of 0 (ie, was never run
        //               successfully, it will surely not be chosen. When this
        //               happens, this argument will be the valid score
        CaseBase( const std::string& ourTeam,
                  const string& fileName,
                  bool writeToFile,
                  bool writeToFileAfterExecution,
                  const double weightThisGame,
                  const double weightThisTeam,
                  const double weightOtherTeams);

        inline ~CaseBase()
        {
            if(writeCasesToFile && !writeToFileAfterExecution)
                writeToFile();
        }

        //Parser
        static CaseBase* parse(const std::string& fileName,
                               const std::string& in,
                               std::string& out);

        unsigned int
        selectSetplay( vector<SetplayID>& possibleSetplayNums,
                       const IWorldModel& world ) const;

        // Evaluates each setplay and inserts the value in the eval slot of
        // SetplayID. Returns evaluation sum.
        double
        evalSetplays(vector<CaseBase::SetplayID>& possibleSetplays,
                     const IWorldModel& world) const;

        void registerCase( const unsigned int setplay_id,
                           const IPoint2D* const ballPos,
                           const PlayMode& gm,
                           bool success,
                           const IWorldModel& world);


        // Output
        void write(std::ostream &os) const;
        void writeToFile() const;


    private:
        typedef std::vector<Case* > Cases;
        typedef std::map<unsigned int, Cases* > SetplayCases;

        //Attributes
        const std::string ourTeam;
        const std::string fileName;
        const bool writeCasesToFile;
        const bool writeToFileAfterExecution;
        const double weightThisGame;
        const double weightThisTeam;
        const double weightOtherTeams;
        SetplayCases M_setplay_cases;
    };

}

std::ostream& operator<<( std::ostream & os,
                          const fcportugal::setplay::Case::Evaluation& e );

#endif //CBR_H
