#include "validator_semantical.h"

#include <sstream>

#include <setplay/model/setplay.h>
#include <setplay/model/transitions.h>
#include <setplay/model/actions.h>
#include <setplay/model/directives.h>
#include <setplay/model/conditions.h>
#include <setplay/model/geom.h>
#include <setplay/model/numbers.h>

namespace setplay
{

    ValidatorSemantical::ValidatorSemantical()
    {

    }

    // FA 03/06 Validate the semantics of the setplay, rather than its syntax => To be called in the end of parsing
    void ValidatorSemantical::visit( Setplay& setplay)
    {
        M_objects.push( &setplay );
        // Create scope for set-play
        /// TODO A parent scope can exist for global variables!!!
        M_scopes.push( Scope( &setplay, "setplay") );

        if ( setplay.hasParameters() )
        {
            Scope& setplay_scope = M_scopes.top();
            for ( Parameters::const_iterator itp = setplay.parameters()->begin();
                  itp != setplay.parameters()->end() ; itp++ )
            {
                setplay_scope.define( new Symbol( (*itp).first,
                                                  (*itp).second->type(),
                                                  &setplay_scope) );
            }
        }

        // Check abort condition of the setplay
        if ( !setplay.hasAbortCondition() )
        {
            stringstream ss;
            ss << "Setplay " << setplay.id()
               << " has no abortCondition. It will never end unless within a step.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            setplay.abortCondition()->visit( *this );
        }

        // Check for redundant step ids
        for ( Steps::const_iterator its1 = setplay.steps()->begin();
              its2 =! setplay.steps()->end() - 1 ; its1++ )
        {
            const Step* s1 = (*it1).second;
            for ( Steps::const_iterator its2 = its1 + 1;
                  its2 =! setplay.steps()->end() ; its2++ )
            {
                const Step* s2 = (*it2).second;

                // Inter-step pairwise semantical validation
                if  ( s1->id() == s2->id() )
                {
                    stringstream ss;
                    ss << "Repeated step id " << s1->id()
                       << " at indexes " << ( its1 - setplay.steps()->begin() )
                       << " and " << ( its2 - setplay.steps()->begin() );
                    addIssue( DuplicateStepId, CRITICAL, ss.str() );
                }
            }
        }

        // Check intra-step validity of each step
        for ( Steps::const_iterator its = setplay.steps()->begin();
              its != setplay.steps()->end(); its++ )
        {
            const Step* step = (*its).second;
            step->visit( *this );

            /// Check that all participants of a step are defined in the set-play
            if ( setplay.players()->size() < step->participations()->size() )
            {
                stringstream ss;
                ss << "Step " << step->id()
                   << " has more participations than the current set-play"
                   << std::endl;
                addIssue( Other, CRITICAL, ss.str() );
            }
        }

        /// Check for unreachable steps

        M_scopes.pop();
        M_objects.pop();
    }

    void ValidatorSemantical::visit( Step& step)
    {
        M_objects.push( &step );
        M_scopes.push( Scope( &step, "step", &M_scopes.top() ) );
        Scope& step_scope = M_scopes.top();

        if ( step.hasAssignments() )
        {
            for ( Assignments::const_iterator ita = step.assignments()->begin();
                  ita != step.assignments()->end(); ita++ )
            {
                step_scope.define( new Symbol( (*ita).first,
                                               (*ita).second->type(),
                                               &step_scope) );
            }
        }

        /// Step AbortCondition
        if ( !step.hasAbortCondition() )
        {
            stringstream ss;
            ss << "Step " << step.id()
               << " has no abort condition. The set-play will never be aborted due to step constraints.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            step.abortCondition()->visit( *this );
        }

        /// Step StartCondition
        if ( !step.hasStartCondition() )
        {
            stringstream ss;
            ss << "Step " << step.id()
               << " has no startCondition. It will start immediatelly after being chosen.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            step.startCondition()->visit( *this );
        }

        /// Step WaitCondition
        if ( !step.hasWaitCondition() )
        {
            stringstream ss;
            ss << "Step " << step.id()
               << " has no waitCondition.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            step.waitCondition()->visit( *this );
        }

        if ( !step.hasLeader() )
        {
            stringstream ss;
            ss << "Step " << step.id()
               << " has no leader specified.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            step.leader()->visit( *this );
        }

        // Check participants of the step
        for ( unsigned int stepParticipantIndex = 0 ;
              stepParticipantIndex < stepParticipations->size();
              stepParticipantIndex++)
        {
            const IParticipation* participation = *itp;
            if ( participation->player()->type() == TT_VAR_PLAYER)
            {
                const VarPlayer* pvar = static_cast<VarPlayer*>( participation->player() );
                if ( step_scope.resolve( pvar->name() ) == NULL )
                {
                    stringstream ss;
                    ss << " Step " << step.id()
                       << " playerVar " << pvar->name()
                       << " is undefined!";
                    addIssue( UndefinedVariable, CRITICAL, ss.str() );
                }
            }
            /* if (!validate(stepParticipations->at(stepParticipantIndex)->player()))
            {
                cerr << "SP_VALID_ERR: Step " << step.id() << " player ";
                stepParticipations->at(stepParticipantIndex)->print(cerr);
                cerr << " doesn't exist as setplay player" << endl;

                valid = false;
            }*/
        }

        // Check transitions of the step
        for ( Transitions::const_iterator itt = step.transitions()->begin();
              itt != step.transitions()->end() ; itt++ )
        {
            if (!validate(step, stepTransitions->at(transitionIndex)))
            {
                cerr << "SP_VALID_ERR: Step " << step.id() << " transition " << transitionIndex << " has an error" << endl;

                valid = false;
            }
        }

        // TODO Are the actions being executed in the transition done by step participants???

        M_scopes.pop();
        M_objects.pop();
    }

    void ValidatorSemantical::visit( IParticipation& participation)
    {

    }

    /// Transitions
    void ValidatorSemantical::visit( NextStep& transition)
    {
        if ( M_setplay->getStep( transition.nextStepId()) == NULL )
        {
            stringstream ss;
            ss << " Step " << transition.nextStepId()
               << " in transition from " << M_step->id()
               << " does not exist.";
            addIssue( TransitionToInexistentStep, CRITICAL, ss.str() );
        }
    }

    void ValidatorSemantical::visit( Finish& transition)
    {
        validate( transition );
    }

    void ValidatorSemantical::visit( Abort& transition)
    {
        validate( transition );
    }

    /// Directives
    void ValidatorSemantical::visit( Do& directive)
    {
        validate( directive );
    }

    void ValidatorSemantical::visit( DoWhile& directive)
    {
        validate( directive );
    }

    void ValidatorSemantical::visit( Dont& directive)
    {
        validate( directive );
    }

    /// Actions
    void ValidatorSemantical::visit( ActDribble& action)
    {
        action.location()->visit( *this );
    }

    void ValidatorSemantical::visit( ActFollow& action)
    {
        action.displacement()->visit( *this );
        action.target()->visit( *this );
    }

    void ValidatorSemantical::visit( ActHold& action )
    {

    }

    void ValidatorSemantical::visit( ActIntercept& action )
    {
        action.target()->visit( *this );
    }

    void ValidatorSemantical::visit( ActPass& action)
    {
        action.location()->visit( *this );
    }

    void ValidatorSemantical::visit( ActPos& action)
    {
        action.location()->visit( *this );
    }

    void ValidatorSemantical::visit( ActShoot& action)
    {

    }

    /// Conditions
    void ValidatorSemantical::visit( CondAnd& condition)
    {
        for ( Conditions::const_iterator itc = condition.conditions()->begin();
              itc != condition.conditions()->end() ; itc++ )
        {
            (*itc)->visit( *this );
        }
    }

    void ValidatorSemantical::visit( CondOr& condition)
    {
        for ( Conditions::const_iterator itc = condition.conditions()->begin();
              itc != condition.conditions()->end() ; itc++ )
        {
            (*itc)->visit( *this );
        }
    }

    void ValidatorSemantical::visit( CondNot& condition)
    {
        condition.condition()->visit( *this );
    }

    void ValidatorSemantical::visit( CondBallOwner& condition)
    {
        condition.player()->visit( *this );
    }

    void ValidatorSemantical::visit( CondTeamInPossession& condition)
    {

    }

    void ValidatorSemantical::visit( CondBool& condition)
    {

    }

    void ValidatorSemantical::visit( CondGameSituation& condition)
    {

    }

    void ValidatorSemantical::visit( CondAtLocation& condition)
    {
        condition.location()->visit( *this );
        condition.object()->visit( *this );
    }

    void ValidatorSemantical::visit( CondTeamFormation& condition)
    {

    }

    void ValidatorSemantical::visit( CondOffside& condition)
    {
        condition.player()->visit( *this );
    }

    void ValidatorSemantical::visit( CondCompDecimal& condition)
    {
        condition.operandLeft()->visit( *this );
        condition.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( CondKickable& condition)
    {
        condition.player()->visit( *this );
    }

    void ValidatorSemantical::visit( CondCanIntercept& condition)
    {
        condition.mobileObject()->visit( *this );
        condition.player()->visit( *this );
    }

    void ValidatorSemantical::visit( CondCanShoot& condition)
    {
        condition.player()->visit( *this );
    }

    void ValidatorSemantical::visit( CondCanPass& condition)
    {
        condition.player()->visit( *this );
    }

    /// Locations
    void ValidatorSemantical::visit( Rect2D& location)
    {
        location.topLeft()->visit( *this );
        location.bottomRight()->visit( *this );
    }

    void ValidatorSemantical::visit( Triangle2D& location)
    {
        location.a()->visit( *this );
        location.b()->visit( *this );
        location.c()->visit( *this );
    }

    void ValidatorSemantical::visit( Sector2D& location)
    {
        location.center()->visit( *this );
        location.startAngle()->visit( *this );
        location.endAngle()->visit( *this );
        location.minRadius()->visit( *this );
        location.maxRadius()->visit( *this );
    }

    void ValidatorSemantical::visit( Circle2D& location)
    {
        location.center()->visit( *this );
        location.radius()->visit( *this );
    }

    void ValidatorSemantical::visit( Ray2D& location)
    {
        location.direction()->visit( *this );
        location.origin()->visit( *this );
    }

    void ValidatorSemantical::visit( Line2D& location)
    {
        location.a()->visit( *this );
        location.b()->visit( *this );
        location.c()->visit( *this );
    }

    /// Participant
    void ValidatorSemantical::visit( PlayerAffiliation& participant)
    {
        // Check the numbers of players
        if ( participant.number() < 1 || participant.number() > 11 )
        {
            stringstream ss;
            ss << "Invalid playerAffiliation "
               << " affiliation= " << participant.affiliation()
               << " number " << participant.number();
            addIssue( Other, CRITICAL, ss.str() );
        }
    }

    /// Player functors
    void ValidatorSemantical::visit( PlayerSelf& player_functor)
    {

    }

    void ValidatorSemantical::visit( NearestPlayer& player_functor)
    {
        player_functor.location()->visit( *this );
    }

    void ValidatorSemantical::visit( NearestTeammate& player_functor)
    {
        player_functor.location()->visit( *this );
    }

    void ValidatorSemantical::visit( NearestOpponent& player_functor)
    {
        player_functor.location()->visit( *this );
    }

    void ValidatorSemantical::visit( Goalie& player_functor)
    {

    }

    /// Points
    void ValidatorSemantical::visit( Point2D& point)
    {
        point.x()->visit( *this );
        point.y()->visit( *this );
    }

    /// Variables => Is this really necessary??
    void ValidatorSemantical::visit( Variable& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "Variable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarLocation& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "LocationVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarRegion& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "RegionVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarPoint& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "PointVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarCondition& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "ConditionVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarDecimal& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "DecimalVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarInteger& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "IntegerVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    void ValidatorSemantical::visit( VarPlayer& variable)
    {
        Scope& scope = M_scopes.top();
        if ( scope.resolve( variable.name() ) )
        {
            stringstream ss;
            ss << "PlayerVariable " << variable.name() << " is undefined!";
            addIssue( Other, Critical, ss.str() );
        }
    }

    /// Point Functors
    void ValidatorSemantical::visit( MobileObjectLocation& point_functor)
    {
        point_functor.mobileObject()->visit( *this );
    }

    void ValidatorSemantical::visit( PlayerPosition& point_functor)
    {
        point_functor.player()->visit( *this );
    }

    void ValidatorSemantical::visit( BallPosition& point_functor)
    {

    }
    // Vector unary operations
    void ValidatorSemantical::visit( VectorNormalize& point_functor)
    {
        point_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorInvertX& point_functor)
    {
        point_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorInvertY& point_functor)
    {
        point_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorInvertXY& point_functor)
    {
        point_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorSubtraction& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorAddition& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    // Vector scalar operations
    void ValidatorSemantical::visit( VectorScale& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorRotation& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorDivision& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( NearestIntersectionPoint& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( RegionCentroid& point_functor)
    {
        point_functor.operandLeft()->visit( *this );
        point_functor.operandRight()->visit( *this );
    }

    /// Simple Data Types
    void ValidatorSemantical::visit( Decimal& number)
    {

    }

    void ValidatorSemantical::visit( Integer& number)
    {

    }

    void ValidatorSemantical::visit( Bool& number )
    {

    }

    /// Number Functors
    // Number operations
    void ValidatorSemantical::visit( NumberAbs& decimal_functor)
    {
        decimal_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( NumberSignChange& decimal_functor)
    {
        decimal_functor.operand()->visit( *this );
    }

    void ValidatorSemantical::visit( NumberMultiplication& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( NumberSubtraction& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( NumberDivision& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( NumberAddition& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    // Vector to scalar operations
    void ValidatorSemantical::visit( VectorDirection& decimal_functor)
    {
        decimal_functor.Left()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorLength& decimal_functor)
    {
        decimal_functor.operand()->visit( *this );
    }

    // Two vector to scalar operations
    void ValidatorSemantical::visit( VectorDistance& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorAngularDiff& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorInnerProduct& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    void ValidatorSemantical::visit( VectorOuterProduct& decimal_functor)
    {
        decimal_functor.operandLeft()->visit( *this );
        decimal_functor.operandRight()->visit( *this );
    }

    // In game queries
    void ValidatorSemantical::visit( DefenseLine& decimal_functor)
    {

    }

    void ValidatorSemantical::visit( OffenseLine& decimal_functor)
    {

    }

    void ValidatorSemantical::visit( OffsideLine& decimal_functor)
    {

    }

    void ValidatorSemantical::visit( MobileObjectSpeed& decimal_functor)
    {
        decimal_functor.mobileObject()->visit( *this );
    }

    void ValidatorSemantical::visit( PointX& decimal_functor )
    {
        decimal_functor.point()->visit( *this );
    }

    void ValidatorSemantical::visit( PointY& decimal_functor )
    {
        decimal_functor.point()->visit( *this );
    }

    /// Integer Functors
    void ValidatorSemantical::visit( CurrentTime& integer_functor)
    {

    }

    void ValidatorSemantical::visit( RemainingTime& integer_functor)
    {

    }

    void ValidatorSemantical::visit( GoalsScored& integer_functor)
    {

    }

    void ValidatorSemantical::visit( GoalDiff& integer_functor)
    {

    }

    void ValidatorSemantical::visit( InterceptionTime& integer_functor )
    {
        integer_functor.mobileObject()->visit( *this );
        integer_functor.player()->visit( *this );
    }

    /// Assignment function
    void ValidatorSemantical::visit( Assignment& assign)
    {
        assign.expression()->visit( *this );
    }

    void ValidatorSemantical::visit( Ball& ball)
    {

    }

    void ValidatorSemantical::validate(const ITransition& transition)
    {
        /// transition startCondition
        if ( !transition.hasStartCondition() )
        {
            stringstream ss;
            ss << "Transition at index " << M_transition_idx
               << " has no startCondition. It will start immediatelly after being chosen.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            transition.startCondition()->visit( *this );
        }

        // Check transition abort condition
        if ( !transition.hasAbortCondition() )
        {
            stringstream ss;
            ss << "Transition at step " << M_transition_idx
               << " has no abortCondition. It will start immediatelly after being chosen.";
            addIssue( Other, WARNING, ss.str() );
        }
        else
        {
            transition.abortCondition()->visit( *this );
        }

        // Check transition directives
        M_directive_idx = 0u;
        for ( Directives::const_iterator itd = transition.directives()->begin();
              itd != transition.directives()->end() ; itd++ )
        {
            (*itd)->visit( *this );
            M_directive_idx++;
        }
    }

    bool ValidatorSemantical::validate(const IDirective& directive)
    {
        // Account for all distinct player references in the directives
        // TODO Create a method that given 2 vectors finds the players that exist in one the don't exist in the other

        for ( Players::const_iterator itp = directive.players()->begin();
              itp != directive.players()->end() ; itp++ )
        {
            if ( !validate(directive->executor()->at(playerIndex), step.participants()) )
            {
                cerr << "SP_VALID_ERR: Directive player ";
                directive->executor()->at(playerIndex)->print(cerr);
                cerr << "SP_VALID_ERR: is not a player of the step!" << endl;
            }
        }

        if (!validate(*transition_directives->at(directiveIndex)))
        {
            cerr << "SP_VALID_ERR: Transition has an invalid directive " << endl;

            valid = false;
        }
        bool valid = true;

        if ( !directive.hasActions() )
        {
            stringstream ss;
            ss << "Directive at index " << M_directive_idx
               << " has no actions defined.";
            addIssue( Other, CRITICAL, ss.str() );
        }
        else
        {
            M_action_idx = 0;
            for ( Actions::const_iterator ita = directive.actions()->begin();
                  ita != directive.actions()->end(); ita++ )
            {
                (*ita)->visit( *this );
                M_action_idx++;
            }
            M_action_idx = -1;
        }

        if (!validate(*directive.executor()))
        {
            valid = false;
        }

    }

    bool ValidatorSemantical::validate(const IPlayer& player, const vector<IParticipation*>* participations)
    {
        // TODO Move Player reference comparison logic to a new method

        for (unsigned int playerIndex = 0; playerIndex < participations->size() ; playerIndex++) {
            if (player->value() == participations->at(playerIndex)->player()) {
                    return true;
            }
        }

        return false;
    }

    void ValidatorSemantical::addIssue( IssueType type,
                                        Severity severity,
                                        std::string message )
    {
        Issue issue;
        issue.type = type;
        issue.severity = severity;
        issue.message = message;
        M_issues.push_back( issue );
    }

    const std::vector<Issue>& ValidatorSemantical::issues() const
    {
        return M_issues;
    }
}

