#include "xml_event_parser.h"

#include <soccer/events/events.h>

// Mandatory for using any feature of Xerces.
#include <xercesc/util/PlatformUtils.hpp>

#include <xercesc/sax2/XMLReaderFactory.hpp>

#include <xercesc/sax2/Attributes.hpp>

#include <soccer/events/xml_event_common.h>

using namespace std::tr1;
XERCES_CPP_NAMESPACE_USE

namespace soccer {
namespace event {

XmlEventParser::XmlEventParser()
{
    XMLPlatformUtils::Initialize();

    M_parser = XMLReaderFactory::createXMLReader();
    M_parser->setContentHandler( this );

}

XmlEventParser::~XmlEventParser()
{

    XMLPlatformUtils::Terminate();
}

void XmlEventParser::parse( const std::string& file_name )
{
    M_parser->parse( file_name.c_str() );
}

const std::string& XmlEventParser::source() const
{
    return M_source;
}


// -----------------------------------------------------------------------
//  Implementations of the SAX DocumentHandler interface
// -----------------------------------------------------------------------
void XmlEventParser::characters( const XMLCh* const chars,
                                 const XMLSize_t length)
{

}

void XmlEventParser::endDocument()
{

}

void XmlEventParser::endElement ( const XMLCh* const uri,
                                  const XMLCh* const localname,
                                  const XMLCh* const qname)
{
    std::string elem_name = S(localname);
    if ( elem_name == "from" || elem_name == "to" )
     {
        return;
    }

    if ( !M_stack.empty() )
    {
        IEvent::EventType event_type = eventType( elem_name );

        shared_ptr<IEvent> last_event = M_stack.top();
        if ( last_event->type() == event_type )
        {
            M_stack.pop();
        }
    }
}

void XmlEventParser::ignorableWhitespace ( const XMLCh* const chars,
                                           const XMLSize_t length )
{

}

void XmlEventParser::processingInstruction ( const XMLCh* const target,
                                             const XMLCh* const data )
{

}

void XmlEventParser::setDocumentLocator(const Locator* const locator)
{

}

void XmlEventParser::startDocument()
{

}

void XmlEventParser::startElement ( const XMLCh* const uri,
                                    const XMLCh* const localname,
                                    const XMLCh* const qname,
                                    const Attributes& attrs )
{
    unsigned int start_time = 0u;
    unsigned int end_time = 0u;
    std::string elem_name( S( localname ) );

    if ( elem_name == "events" )
    {
        const XMLCh* source_value_ch = attrs.getValue( X("source") );
        if ( source_value_ch != NULL )
        {
            M_source = std::string( S(source_value_ch) );
        }
        // Ignore events element
        return;
    }

    if ( elem_name == "player" )
    {
        shared_ptr<IEvent> last_event = M_events.back();
        // TODO HACK !!! Check type of last event quickly

        // Continue the instantiation of a previous event
        switch ( last_event->type() )
        {
            case IEvent::BallDispute:
            {
                BallDispute* ball_dispute = static_cast< BallDispute* >( last_event.get() );

                /// Parse a player element
                TeamSide team_side = TS_UNKNOWN;
                parseTeamSideAttribute( attrs, "teamSide", team_side);
                unsigned int number = 0u;
                parsePlayerNumberAttribute( attrs, "number", number );

                ball_dispute->addPlayer( PlayerId( number, team_side ) );
            }
            break;
            default:
            {
                std::cerr << "XmlEventParser: Unhandled/Illegal event type "
                          << eventType( last_event->type() )  << " continued instantiation " << std::endl;
                assert(false);
            }
            break;
        }
        return;
    }

    if ( elem_name == "from" )
    {
        rcsc::Vector2D vector;
        parseVector( attrs, vector );
        shared_ptr<IEvent> event = M_stack.top();

        switch ( event->type() )
        {
            case IEvent::BallMove:
            {
                BallMove* ball_move = static_cast<BallMove*>( event.get() );
                ball_move->setFrom( vector );
            }
            break;
            case IEvent::PlayerMove:
            {
                PlayerMove* player_move = static_cast<PlayerMove*>( event.get() );
                player_move->setFrom( vector );
            }
            break;
            default:
                assert(false);
                break;
        }

        return;
    }

    if ( elem_name == "to" )
    {
        rcsc::Vector2D vector;
        parseVector( attrs, vector );
        shared_ptr<IEvent> event = M_stack.top();

        switch ( event->type() )
        {
            case IEvent::BallMove:
            {
                BallMove* ball_move = static_cast<BallMove*>( event.get() );
                ball_move->setTo( vector );
            }
            break;
            case IEvent::PlayerMove:
            {
                PlayerMove* player_move = static_cast<PlayerMove*>( event.get() );
                player_move->setTo( vector );
            }
            break;
            default:
                assert(false);
                break;
        }

        return;
    }

    IEvent::EventType event_type = eventType( elem_name );

    // Parse individual event
    parseEventBaseAttributes( attrs, start_time, end_time);

    shared_ptr<IEvent> new_event;

    switch ( event_type )
    {
        case IEvent::KickOff:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new KickOff( start_time, team_side ));
        }
        break;
        case IEvent::KickOffTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new KickOffTaken( start_time, team_side ));
        }
        break;
        case IEvent::ThrowIn:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new ThrowIn( start_time, team_side ));
        }
        break;
        case IEvent::ThrowInTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new ThrowInTaken( start_time, team_side ));
        }
        break;
        case IEvent::GoalKick:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new GoalKick( start_time, team_side ));
        }
        break;
        case IEvent::GoalKickTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new GoalKickTaken( start_time, team_side ));
        }
        break;
        case IEvent::CornerKick:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new CornerKick( start_time, team_side ));
        }
        break;
        case IEvent::CornerKickTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new CornerKickTaken( start_time, team_side ));
        }
        break;
        case IEvent::FreeKick:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new FreeKick( start_time, team_side ));
        }
        break;
        case IEvent::FreeKickTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new FreeKickTaken( start_time, team_side ));
        }
        break;
        case IEvent::IndirectFreeKick:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new IndirectFreeKick( start_time, team_side ));
        }
        break;
        case IEvent::IndirectFreeKickTaken:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );

            new_event = shared_ptr< IEvent> ( new IndirectFreeKickTaken( start_time, team_side ));
        }
        break;
        case IEvent::SoccerMomentChanged:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side );
            const XMLCh* moment_ch = attrs.getValue( X("moment") );
            assert( moment_ch != NULL );
            SoccerMomentChanged::TeamSideMoment::Moment moment = SoccerMomentChanged::TeamSideMoment::moment( std::string( S( moment_ch ) ) );

            new_event = shared_ptr< IEvent> ( new SoccerMomentChanged( start_time,
                                                 SoccerMomentChanged::TeamSideMoment( team_side,
                                                                                      moment) ));
        }
        break;
        case IEvent::PossessionChanged:
        {
            TeamSide team_in_possession = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamInPossession", team_in_possession  );

            new_event = shared_ptr< IEvent> ( new PossessionChanged( start_time, team_in_possession ));
        }
        break;
        case IEvent::Interception:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side   );
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr< IEvent> ( new Interception( start_time, PlayerId( number, team_side ) ));
        }
        break;
        case IEvent::Turnover:
        {
            TeamSide team_in_possession = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamInPossession", team_in_possession  );

            new_event = shared_ptr< IEvent> ( new Turnover( start_time, team_in_possession ));
        }
        break;
        case IEvent::BallDispute:
        {
            new_event = shared_ptr< IEvent> ( new BallDispute( start_time ));
        }
        break;
        case IEvent::BallHold:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr<IEvent>( new BallHold( start_time, end_time,
                                                          PlayerId( number, team_side )));
        }
        break;
        case IEvent::GoalieCatch:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);

            new_event = shared_ptr< IEvent> ( new GoalieCatch( start_time, team_side ));
        }
        break;
        case IEvent::BallKick:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr< IEvent> ( new BallKick( start_time,
                                                            PlayerId( number, team_side )));
        }
        break;
        case IEvent::Goal:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);

            new_event = shared_ptr< IEvent> ( new Goal( start_time, team_side ));
        }
        break;
        case IEvent::Shoot:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr< IEvent> ( new Shoot( start_time, PlayerId( number, team_side)) );
        }
        break;
        case IEvent::Dribble:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr< IEvent> ( new Dribble( start_time, end_time,
                                                           PlayerId( number, team_side)));
        }
        break;
        case IEvent::PlayerMove:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int number = 0u;
            parsePlayerNumberAttribute( attrs, "number", number );

            new_event = shared_ptr< IEvent> ( new PlayerMove( start_time, end_time,
                                                              PlayerId( number, team_side)));
            M_stack.push( new_event );
            M_events.push_back( new_event );
            return;
        }
        break;
        case IEvent::BallMove:
        {
            new_event = shared_ptr< IEvent >( new BallMove( start_time,
                                                            end_time) );
            M_stack.push( new_event );
            M_events.push_back( new_event );
            return;
        }
        break;
        case IEvent::Pass:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);
            unsigned int receiver = 0u;
            parsePlayerNumberAttribute( attrs, "receiver", receiver );
            unsigned int sender = 0u;
            parsePlayerNumberAttribute( attrs, "sender", sender );

            new_event = shared_ptr< IEvent> ( new Pass( start_time, end_time,
                                                        team_side,
                                                        sender, receiver));
        }
            break;
        case IEvent::Offside:
        {
            TeamSide team_side = TS_UNKNOWN;
            parseTeamSideAttribute( attrs, "teamSide", team_side);

            new_event = shared_ptr< IEvent> ( new Offside( start_time, team_side));
        }
        break;
        case IEvent::PassChain:
        {
            // TODO Handle pass-chain events
            new_event = shared_ptr< IEvent> ( new PassChain( ) );
            if ( M_stack.empty() )
            {
                M_events.push_back( new_event );
            }
            else
            {
                EventChain* event_chain = static_cast<EventChain*>( M_stack.top().get() );
                event_chain->push_back( new_event );
            }
            M_stack.push(new_event);
            return;
        }
        break;
        case IEvent::GoalPassChain:
        {
            // TODO Handle pass-chain events
            new_event = shared_ptr< IEvent> ( new GoalPassChain( ) );
            M_stack.push( new_event );
            M_events.push_back( new_event );
            return;
        }
        break;
        default:
        {
            std::cerr << "XmlEventParser: Unhandled/Illegal event type " << event_type << std::endl;
            assert(false);
        }
        break;
    }

    if ( !M_stack.empty() )
    {
        /// Get the a reference to the last element inserted
        shared_ptr<IEvent> last_event = M_stack.top();
        switch ( last_event->type() )
        {
            case IEvent::PassChain:
            case IEvent::GoalPassChain:
            {
                EventChain* event_chain = static_cast< EventChain* >( last_event.get() );
                event_chain->push_back( new_event );
            }
            break;
            default:
            {
                std::cerr << "XmlEventParser: Unexpected element at the top of the stack "
                          << eventType( last_event->type() )
                          << std::endl;
                assert( false );
            }
            break;
        }
        return;
    }

    M_events.push_back( new_event );
}

void XmlEventParser::startPrefixMapping( const XMLCh* const prefix,
                                         const XMLCh* const uri )
{

}

void XmlEventParser::endPrefixMapping( const XMLCh* const prefix)
{

}

void XmlEventParser::skippedEntity( const XMLCh* const name )
{
    std::cerr << "Skipped parsing " << name << std::endl;
}

bool XmlEventParser::parseEventBaseAttributes( const Attributes& attrs,
                                               unsigned int& start_time,
                                               unsigned int& end_time )
{
    const XMLCh* start_time_ch = attrs.getValue( X("startTime" ));
    assert( start_time_ch != NULL );
    start_time = atoi( S(start_time_ch) );

    const XMLCh* end_time_ch = attrs.getValue( X("endTime" ));
    if ( end_time_ch != NULL )
    {
        end_time = atoi( S(end_time_ch) );
    }
    else
    {
        end_time = start_time;
        return false;
    }
    return true;
}

bool XmlEventParser::parseTeamSideAttribute( const Attributes& attrs,
                                             const char* attr_name,
                                             TeamSide& team_side )
{
    const XMLCh* team_side_ch = attrs.getValue( X(attr_name) );

    if ( team_side_ch == NULL )
    {
        return false;
    }

    team_side = teamSide( S( team_side_ch ) );

    return true;
}

bool XmlEventParser::parsePlayerNumberAttribute( const Attributes& attrs,
                                                 const char* attr_name,
                                                 unsigned int& number)
{
    const XMLCh* number_ch = attrs.getValue( X(attr_name) );

    if ( number_ch == NULL )
    {
        return false;
    }

    number = atoi( S( number_ch ) );

    return true;
}

bool XmlEventParser::parseVector( const xercesc::Attributes& attrs,
                                  rcsc::Vector2D& vector)
{
    const XMLCh* coordx_ch = attrs.getValue( X("x") );
    if ( coordx_ch == NULL )
    {
        return false;
    }

    float x = atof( S( coordx_ch ) );

    const XMLCh* coordy_ch = attrs.getValue( X("y") );
    if ( coordy_ch == NULL )
    {
        return false;
    }

    float y = atof( S( coordy_ch ) );

    vector.assign( x, y );

    return true;
}

}
}
