#include "flux_dt.h"

#include <soccer/iworld_model.h>

using namespace rcsc;
using namespace soccer::strategy::util;

namespace soccer {
namespace flux {

    FluxDT::FluxDT(const rcsc::Rect2D& bounding_rect)
        : M_flux_values( NULL)
    {
        M_triangulation.init( bounding_rect );
    }


    FluxDT::~FluxDT()
    {
        if ( M_flux_values != NULL )
        {
            delete [] M_flux_values;
        }
    }


    FluxType FluxDT::type() const
    {
        return FT_Delaunay;
    }


    double FluxDT::value(const rcsc::Vector2D& position) const
    {
        const DelaunayTriangulation::Triangle* dt_triangle = M_triangulation.findTriangleContains(position);

        if ( dt_triangle == NULL )
        {
            dlog.addText ( Logger::POSITIONING,
                          __FILE__ " Triangle not found for position (%.2f,%.2f)",
                          position.x, position.y);
            const int nearest_index = M_triangulation.findNearestVertex(position)->id();
            return M_flux_values[nearest_index];
        }

        if ( position == M_triangulation.getVertex(dt_triangle->vertex(0)->id())->pos() )
        {
            return M_flux_values[dt_triangle->vertex(0)->id()];
        }
        else if ( position == M_triangulation.getVertex(dt_triangle->vertex(1)->id())->pos() )
        {
            return M_flux_values[dt_triangle->vertex(1)->id()];
        }
        else if ( position == M_triangulation.getVertex(dt_triangle->vertex(2)->id())->pos() )
        {
            return M_flux_values[dt_triangle->vertex(2)->id()];
        }

        return interpolate(position,
                           M_triangulation.getVertex(dt_triangle->vertex(0)->id())->pos(),
                           M_triangulation.getVertex(dt_triangle->vertex(1)->id())->pos(),
                           M_triangulation.getVertex(dt_triangle->vertex(2)->id())->pos(),
                           Vector2D(M_flux_values[dt_triangle->vertex(0)->id()], 0.0),
                           Vector2D(M_flux_values[dt_triangle->vertex(1)->id()], 0.0),
                           Vector2D(M_flux_values[dt_triangle->vertex(2)->id()], 0.0)).x;
    }


    bool FluxDT::read(std::istream& is)
    {
        std::cout << "Reading Delaunay Flux File" << std::endl;
        //TODO confirm that file is of correct type (2D, 3D or MR)

        std::istringstream istr(getNextLine(is));

        // BEGIN FLUX NUMBER_OF_SAMPLES

        unsigned int num_samples = 0;

        if ( sscanf( istr.str().c_str(), " Begin Flux %u ", &num_samples ) != 1 )
        {
            std::cerr << "FluxDelaunay: Error failed to read flux data size." << std::endl;
            return false;
        }

        M_flux_values = new double[num_samples];
        for ( unsigned int i = 0; i < num_samples; i++ )
        {
            // Get the next line of data
            istr.clear();
            istr.str(getNextLine(is));

            unsigned int index;
            double ball_pos_x, ball_pos_y;
            if ( sscanf( istr.str().c_str(),
                        " %u %lf %lf %lf ",
                        &index, &ball_pos_x, &ball_pos_y, &M_flux_values[i] ) != 4 || index != i )
            {
                std::cerr << "FluxDelaunay: Error failed to read flux data." << std::endl;
                return false;
            }
            M_triangulation.addVertex(ball_pos_x, ball_pos_y);
        }

        // Compute the triangulation after importing the
        M_triangulation.compute();
        std::cout << "Finish Reading Delaunay Flux File " << std::endl;
        return true;
    }



    void FluxDT::update(const IWorldModel* )
    {

    }

}
}
